"""Shared configuration from which templates can be loaded and parsed."""

from __future__ import annotations
from pathlib import Path

from typing import Callable
from typing import Dict
from typing import Any
from typing import Type
from typing import Union
from typing import Optional
from typing import Mapping

import warnings

from liquid.mode import Mode
from liquid.tag import Tag
from liquid.template import Template
from liquid.lex import get_lexer
from liquid.stream import TokenStream
from liquid.parse import get_parser

from liquid import ast
from liquid import loaders

from liquid.exceptions import Error
from liquid.exceptions import lookup_warning


# pylint: disable=too-many-instance-attributes
class Environment:
    """Shared configuration from which templates can be loaded and parsed.

    An ``Environment`` is also where you'd register custom tags and filters.

    Args:
        tag_start_string: The sequence of characters indicating the start of a
            liquid tag.
        tag_end_string: The sequence of characters indicating the end of a
            liquid tag.
        statement_start_string: The sequence of characters indicating the start
            of an output statement.
        statement_end_string: The sequence of characters indicating the end
            of an output statement.
        strip_tags: If `True` will strip preceding and trailing whitespace from
            all tags, regardless of whitespace control characters.
        tolerance: Indicates how tolerant to be of errors. Must be one of
            `Mode.LAX`, `Mode.WARN` or `Mode.STRICT`.
        loader: A template loader. If you want to use the builtin "render" or
            "include" tags, a loader must be configured.
        globals: A mapping that will be added to the context of any template
            loaded from this environment.
    """

    # pylint: disable=redefined-builtin too-many-arguments
    def __init__(
        self,
        tag_start_string: str = r"{%",
        tag_end_string: str = r"%}",
        statement_start_string: str = r"{{",
        statement_end_string: str = r"}}",
        strip_tags: bool = False,
        tolerance: Mode = Mode.STRICT,
        loader: Optional[loaders.BaseLoader] = None,
        globals: Optional[Mapping[str, object]] = None,
    ):
        self.tag_start_string = tag_start_string
        self.tag_end_string = tag_end_string
        self.statement_start_string = statement_start_string
        self.statement_end_string = statement_end_string
        self.strip_tags = strip_tags
        self.loader = loader or loaders.DictLoader({})
        self.globals: Mapping[str, object] = globals or {}

        # Tag register.
        self.tags: Dict[str, Tag] = {}

        # Filter register.
        self.filters: Dict[str, Callable[..., Any]] = {}

        # tolerance mode
        self.mode = tolerance

        # Template cache
        from liquid.utils import LRUCache

        self.cache = LRUCache(300)

        from liquid import builtin

        builtin.register(self)

    def __hash__(self):
        return hash(
            (
                self.statement_start_string,
                self.statement_end_string,
                self.tag_start_string,
                self.tag_end_string,
                self.mode,
                self.strip_tags,
            )
        )

    def add_tag(self, tag: Type[Tag]) -> None:
        """Register a liquid tag with the environment."""
        self.tags[tag.name] = tag(self)

    def add_filter(self, name: str, func: Callable[..., Any]) -> None:
        """Register a filter function with the environment.

        Args:
            name: The filter's name. Does not need to match the function name.
                This is what you'll use to apply the filter to an expression in
                a liquid template.
            func: Any callable that accepts at least one argument, the result
                of the expression the filter is applied to.

                If the filter needs access to the environment or render
                context, you probably want to make `func` a class that inherits
                from `liquid.Filter`, and override the `__call__` method. All
                builtin filters are implemented in this way.
        """
        self.filters[name] = func

    def parse(self, source: str) -> ast.ParseTree:
        """Parse the given string as a liquid template.

        More often than not you'll want to use `Environment.from_string` instead.
        """
        tokenize = get_lexer(
            self.tag_start_string,
            self.tag_end_string,
            self.statement_start_string,
            self.statement_end_string,
        )
        parser = get_parser(self)

        token_iter = tokenize(source)
        parse_tree = parser.parse(TokenStream(token_iter))
        return parse_tree

    # pylint: disable=redefined-builtin
    def from_string(
        self,
        source: str,
        name: str = "",
        path: Optional[Path] = None,
        globals: Optional[Mapping[str, object]] = None,
    ) -> Template:
        """Parse the given string as a liquid template.

        `name` and `path` are passed to the `Template` constructor and will be
        used to populate the template's `TemplateDrop`.

        Args:
            source: Liquid template source.
            name: Name of the template.
            path: Location of the template.
            globals: A mapping of context variables made available every time
                the template is rendered.
        Returns:
            A parsed template ready to be rendered.
        """
        parse_tree = self.parse(source)
        return Template(
            env=self,
            name=name,
            path=path,
            parse_tree=parse_tree,
            globals=self.make_globals(globals),
        )

    # pylint: disable=redefined-builtin
    def get_template(
        self, name: str, globals: Optional[Mapping[str, object]] = None
    ) -> Template:
        """Load and parse a template using the configured loader.

        Args:
            name: The template's name. The loader is responsible for
                interpretting the name.
            globals: A mapping of context variables made available every time
                the resulting template is rendered.
        Returns:
            A parsed template ready to be rendered.
        """
        template = self.cache.get(name)

        if isinstance(template, Template) and template.is_up_to_date:
            # Copy the cached template with new globals.
            return Template(
                env=self,
                name=template.name,
                path=template.path,
                parse_tree=template.tree,
                globals=self.make_globals(globals),
            )

        template = self.loader.load(self, name, globals=self.make_globals(globals))
        self.cache[name] = template
        return template

    # pylint: disable=redefined-builtin
    def make_globals(
        self, globals: Optional[Mapping[str, object]] = None
    ) -> Mapping[str, object]:
        """Combine environment globals with template globals."""
        if globals:
            _globals: Dict[str, object] = {**self.globals, **globals}
        else:
            _globals: Dict[str, object] = {**self.globals}
        return _globals

    def error(
        self,
        exc: Union[Type[Error], Error],
        msg: Optional[str] = None,
        linenum: Optional[int] = None,
    ) -> None:
        if not isinstance(exc, Error):
            exc = exc(msg, linenum=linenum)
        elif not exc.linenum:
            exc.linenum = linenum

        if self.mode == Mode.STRICT:
            raise exc
        if self.mode == Mode.WARN:
            warnings.warn(str(exc), category=lookup_warning(exc.__class__))



"""Filter function helpers."""

import collections.abc

from abc import ABC, abstractmethod
from functools import wraps

from typing import Tuple
from typing import Any
from typing import Protocol

from liquid.mode import Mode
from liquid.exceptions import FilterArgumentError


class Env(Protocol):
    """Environment interface expected by `Filter`"""

    mode: Mode


class Filter(ABC):
    """Base class for all filters.

    If with_context is True, the current render context will be passed as a
    keyword argument every time the filter is called.
    """

    __slots__ = ("env",)

    with_context = False

    def __init__(self, env: Env):
        self.env = env

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """Call the filter where `val` is the left hand expression of an output statement."""


def expect_number(name: str, val: object):
    """Return `val` as an integer or float. Raise a FilterArgumentError
    if `val` can't be cast to an integer or float."""
    if isinstance(val, (int, float)):
        num = val
    elif isinstance(val, str):
        if val.isdecimal():
            num = int(val)
        try:
            num = float(val)
        except ValueError as err:
            raise FilterArgumentError(
                f"{name}: could not cast string '{val}' to a number"
            ) from err
    else:
        raise FilterArgumentError(
            f"expected an int, float or string, found {type(val)}"
        )
    return num


def expect_integer(name: str, val: object):
    """Return `val` as an integer. Raise a FilterArgumentError if `val` can't
    be cast to an integer."""
    if isinstance(val, int):
        num = val
    elif isinstance(val, str):
        if val.isdecimal():
            num = int(val)
        else:
            raise FilterArgumentError(
                f"{name}: could not cast string '{val}' to an integer"
            )
    else:
        raise FilterArgumentError(f"expected an int or string, found {type(val)}")
    return num


def expect_string(name: str, val: object):
    """Raise a FilterArgumentError is val is not a string."""
    if not isinstance(val, str):
        raise FilterArgumentError(f"{name}: expected a string, found {type(val)}")


def expect_string_or_array(name: str, val: object):
    """Raise a FilterArgumentError is val is not a string or a list."""
    if not isinstance(val, (str, list)):
        raise FilterArgumentError(
            f"{name}: expected a string or array, found {type(val)}"
        )


def expect_n_args(name: str, n: int, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args does not have `n` elements, and
    kwargs is not empty."""
    length = len(args)
    if length != n:
        raise FilterArgumentError(
            f"{name} expects a one positional argument, found {length}"
        )
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )


def expect_one_arg(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args is not a single element tuple, and
    kwargs is not empty."""
    expect_n_args(name, 1, args, kwargs)


def maybe_one_arg(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args is not empty or a single element tuple,
    and kwargs is not empty."""
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )

    length = len(args)
    if length == 1:
        return True
    if length == 0:
        return False

    raise FilterArgumentError(
        f"{name}: expected zero or one positional arguments, found {length}"
    )


def one_maybe_two_args(name: str, args: ..., kwargs: ...) -> Tuple[Any, ...]:
    """Return (args[0], args[1]) or None if the args aren't there."""
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )

    length = len(args)
    if length == 2:
        return args[0], args[1]
    if length == 1:
        return args[0], None

    raise FilterArgumentError(
        f"{name}: expected one or two positional arguments, found {length}"
    )


def expect_no_args(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args and kwargs are not empty."""
    if args:
        raise FilterArgumentError(
            f"{name}: expected zero positional arguments, found {len(args)}"
        )
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )


def expect_array(name: str, val: object):
    """Raise a FilterArgumentError is val is not an array/list."""
    if not isinstance(val, (list, tuple)):
        raise FilterArgumentError(f"{name}: expected an array, found {type(val)}")


def array_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_or_string_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_string_or_array(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_of_strings_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        for item in val:
            if not isinstance(item, str):
                raise FilterArgumentError(
                    f"{instance.name}: expected an array of strings, found a {type(item)}"
                )
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_of_hashable_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        for item in val:
            if not isinstance(item, collections.abc.Hashable):
                raise FilterArgumentError(
                    f"{instance.name}: expected an array of hashable items, "
                    f"found a {type(item)}"
                )
        return func(instance, val, *args, **kwargs)

    return wrapper


def string_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_string(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def number_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        num = expect_number(instance.name, val)
        return func(instance, num, *args, **kwargs)

    return wrapper


def one_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        return func(instance, val, args[0], **kwargs)

    return wrapper


def maybe_one_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        maybe_one_arg(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper


def one_maybe_two_args_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        one_maybe_two_args(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper


def one_array_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        expect_array(instance.name, args[0])
        return func(instance, val, args[0], **kwargs)

    return wrapper


def one_number_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        num = expect_number(instance.name, args[0])
        return func(instance, val, num, **kwargs)

    return wrapper


def one_string_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        expect_string(instance.name, args[0])
        return func(instance, val, args[0], **kwargs)

    return wrapper


def two_string_args_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_n_args(instance.name, 2, args, kwargs)
        expect_string(instance.name, args[0])
        expect_string(instance.name, args[1])
        return func(instance, val, args[0], args[1], **kwargs)

    return wrapper


def no_args(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_no_args(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper



"""Test math filter functions."""
import unittest
from inspect import isclass
from typing import NamedTuple, Any, List, Dict

from liquid.environment import Environment
from liquid.exceptions import FilterArgumentError

from liquid.builtin.filters.array import (
    Join,
    First,
    Last,
    Concat,
    Map,
    Reverse,
    Sort,
    SortNatural,
    Where,
    Uniq,
    Compact,
)


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class ArrayFilterTestCase(unittest.TestCase):
    """Test array filter functions."""

    def _test(self, filter_cls, test_cases):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        func = filter_cls(env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, FilterArgumentError
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_join(self):
        """Test `join` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[
                    ", ",
                ],
                kwargs={},
                expect="a, b",
            ),
            Case(
                description="missing argument",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=["a", "b"],
                args=[", ", ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arguments not a string",
                val=["a", "b"],
                args=[5],
                kwargs={},
                expect="a5b",
            ),
            Case(
                description="value not an array",
                val=12,
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value array contains non string",
                val=["a", "b", 5],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Join, test_cases)

    def test_first(self):
        """Test `first` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect="a",
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect="a",
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(First, test_cases)

    def test_last(self):
        """Test `last` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect="b",
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect={},
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=None,
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Last, test_cases)

    def test_concat(self):
        """Test `concat` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b"],
                args=[["c", "d"]],
                kwargs={},
                expect=["a", "b", "c", "d"],
            ),
            Case(
                description="missing argument",
                val=["a", "b"],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=["a", "b"],
                args=[["c", "d"], ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arguments not a list",
                val=["a", "b"],
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not an array",
                val="a, b",
                args=[["c", "d"]],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="array contains non string",
                val=["a", "b", 5],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Concat, test_cases)

    def test_map(self):
        """Test `map` filter function."""

        test_cases = [
            Case(
                description="lists of objects",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title"],
                kwargs={},
                expect=["foo", "bar", "baz"],
            ),
            Case(
                description="missing argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title", ""],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing property",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "baz"}],
                args=["title"],
                kwargs={},
                expect=["foo", "bar", None],
            ),
            Case(
                description="value not an array",
                val=123,
                args=["title"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="array contains non object",
                val=[{"title": "foo"}, {"title": "bar"}, 5, []],
                args=["title"],
                kwargs={},
                expect=["foo", "bar", None, None],
            ),
        ]

        self._test(Map, test_cases)

    def test_reverse(self):
        """Test `reverse` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "B", "A"],
                args=[],
                kwargs={},
                expect=["A", "B", "a", "b"],
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, [], {}],
                args=[],
                kwargs={},
                expect=[{}, [], 1, "b", "a"],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=123,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Reverse, test_cases)

    def test_sort(self):
        """Test `sort` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "C", "B", "A"],
                args=[],
                kwargs={},
                expect=["A", "B", "C", "a", "b"],
            ),
            Case(
                description="lists of objects with key",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "Baz"}, {"title": "bar"}, {"title": "foo"}],
            ),
            Case(
                description="lists of objects with missing key",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "foo"}, {"heading": "Baz"}],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "baz"}],
                args=["title", "heading"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=123,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Sort, test_cases)

    def test_sort_natural(self):
        """Test `sort_natural` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["b", "a", "C", "B", "A"],
                args=[],
                kwargs={},
                expect=["a", "A", "b", "B", "C"],
            ),
            Case(
                description="lists of strings with a None",
                val=["b", "a", None, "C", "B", "A"],
                args=[],
                kwargs={},
                expect=[None, "a", "A", "b", "B", "C"],
            ),
            Case(
                description="lists of objects with key",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "Baz"}, {"title": "foo"}],
            ),
            Case(
                description="lists of objects with missing key",
                val=[{"title": "foo"}, {"title": "bar"}, {"heading": "Baz"}],
                args=["title"],
                kwargs={},
                expect=[{"title": "bar"}, {"title": "foo"}, {"heading": "Baz"}],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}],
                args=["title", "heading"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1234,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(SortNatural, test_cases)

    def test_where(self):
        """Test `where` filter function."""

        test_cases = [
            Case(
                description="lists of object",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title"],
                kwargs={},
                expect=[{"title": "foo"}, {"title": "bar"}],
            ),
            Case(
                description="lists of object with equality test",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title", "bar"],
                kwargs={},
                expect=[{"title": "bar"}],
            ),
            Case(
                description="value not an array",
                val=1234,
                args=["title"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val=[{"title": "foo"}, {"title": "bar"}, {"title": None}],
                args=["title", "bar", "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Where, test_cases)

    def test_uniq(self):
        """Test `uniq` filter function."""

        test_cases = [
            Case(
                description="lists of strings",
                val=["a", "b", "b", "a"],
                args=[],
                kwargs={},
                expect=["a", "b"],
            ),
            Case(
                description="lists of things",
                val=["a", "b", 1, 1],
                args=[],
                kwargs={},
                expect=["a", "b", 1],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unhashable items",
                val=["a", "b", [], {}],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val="a, b",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Uniq, test_cases)

    def test_compact(self):
        """Test `compact` filter function."""

        test_cases = [
            Case(
                description="lists with nil",
                val=["b", "a", None, "A"],
                args=[],
                kwargs={},
                expect=["b", "a", "A"],
            ),
            Case(
                description="empty list",
                val=[],
                args=[],
                kwargs={},
                expect=[],
            ),
            Case(
                description="unexpected argument",
                val=["a", "b"],
                args=[", "],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="value not an array",
                val=1,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Compact, test_cases)



"""Liquid specific Excpeptions and wanrings."""

from typing import Dict
from typing import Type
from typing import Optional
from typing import Union

from pathlib import Path


class Error(Exception):
    def __init__(
        self,
        *args: object,
        linenum: Optional[int] = None,
        filename: Optional[Union[str, Path]] = None,
    ):
        self.linenum = linenum
        self.filename = filename
        super().__init__(*args)

    def __str__(self):
        msg = super().__str__()
        if self.linenum:
            msg = f"{msg}, on line {self.linenum}"
        if self.filename:
            msg += f" of {self.filename}"
        return msg


class LiquidInterrupt(Exception):
    """Loop interrupt"""


class LiquidSyntaxError(Error):
    """Exception raised when there is a parser error."""


class LiquidTypeError(Error):
    """Exception raised when an error occurs at render time."""


class LiquidKeyError(Error):
    """Exception raised when an undefined variable is accessed."""


class DisabledTagError(Error):
    """Exception raised when an attempt is made to render a disabled tag."""


class NoSuchFilterFunc(Error):
    """Exception raised when a filter lookup fails."""


class FilterArgumentError(Error):
    """Exception raised when a filters arguments are invalid."""


class TemplateNotFound(Error):
    """Excpetions raised when a template could not be found."""

    def __str__(self):
        msg = super().__str__()
        msg = f"template not found {msg}"
        return msg


class ContextDepthError(Error):
    """Exception raised when the maximum context depth is reached."""


class BreakLoop(LiquidInterrupt):
    """Exception raised when a BreakNode is rendered."""


class ContinueLoop(LiquidInterrupt):
    """Exception raised when a ContinueNode is rendered."""


class LiquidWarning(UserWarning):
    """Base warning."""


class LiquidSyntaxWarning(LiquidWarning):
    """Replaces LiquidSyntaxError when in WARN mode."""


class LiquidTypeWarning(LiquidWarning):
    """Replaces LiquidTypeError when in WARN mode."""


class FilterWarning(LiquidWarning):
    """Replaces filter exceptions when in WARN mode."""


class OpcodeError(Error):
    """Exceptions raised when there is an opcode definition error."""


class StackOverflow(Error):
    """Exception raised when the stack is full."""


WARNINGS: Dict[Type[Error], Type[LiquidWarning]] = {
    LiquidSyntaxError: LiquidSyntaxWarning,
    LiquidTypeError: LiquidTypeWarning,
    FilterArgumentError: FilterWarning,
    NoSuchFilterFunc: FilterWarning,
}


def lookup_warning(exc: Type[Error]) -> Type[LiquidWarning]:
    return WARNINGS.get(exc, LiquidWarning)



import json

from liquid.filter import Filter, no_args
from liquid.exceptions import FilterArgumentError

# pylint: disable=arguments-differ, too-few-public-methods


class JSON(Filter):

    __slots__ = ()

    name = "json"

    @no_args
    def __call__(self, mapping):
        if not isinstance(mapping, dict):
            raise FilterArgumentError(
                f"{self.name}: expected an object, found {type(mapping)}"
            )
        return json.dumps({k: v for k, v in mapping.items() if k != "collections"})



__version__ = "0.5.2"

from liquid.mode import Mode
from liquid.filter import Filter
from liquid.token import Token
from liquid.context import Context
from liquid.expression import Expression
from liquid.environment import Environment
from liquid.environment import Template
from liquid.loaders import FileSystemLoader



"""Tag and output statement expression objects."""

from __future__ import annotations

from abc import ABC
from abc import abstractmethod

from collections import abc
from itertools import islice

from typing import Union
from typing import List
from typing import Optional
from typing import Any
from typing import Iterator
from typing import Tuple
from typing import Generic
from typing import TypeVar

from liquid.context import Context
from liquid.exceptions import LiquidTypeError, Error

# pylint: disable=missing-class-docstring too-few-public-methods


class Expression(ABC):
    __slots__ = ()

    expression_statement = True

    @abstractmethod
    def evaluate(self, context: Context) -> object:
        """Evaluate the expression with the given context."""


class Nil(Expression):
    __slots__ = ()

    def __eq__(self, other: object):
        return isinstance(other, Nil)

    def __repr__(self):  # pragma: no cover
        return "Nil()"

    def __str__(self):  # pragma: no cover
        return "nil"

    def evaluate(self, context: Context) -> None:
        return None


NIL = Nil()


class Empty(Expression):
    __slots__ = ()

    def __eq__(self, other: object):
        if isinstance(other, Empty):
            return True
        if isinstance(other, (list, dict, str)) and not other:
            return True
        return False

    def __repr__(self):  # pragma: no cover
        return "Empty()"

    def __str__(self):  # pragma: no cover
        return "empty"

    def evaluate(self, context: Context) -> Empty:
        return self


EMPTY = Empty()


T = TypeVar("T")  # pylint: disable=invalid-name


class Literal(Expression, Generic[T]):
    __slots__ = "value"

    def __init__(self, value: T):
        self.value = value

    def __str__(self):
        return repr(self.value)

    def __hash__(self):
        return hash(self.value)

    def evaluate(self, context: Context) -> object:
        return self.value


class Boolean(Literal):
    __slots__ = ()

    def __init__(self, value: bool):
        super().__init__(value)

    def __eq__(self, other: object):
        return isinstance(other, Boolean) and self.value == other.value

    def __repr__(self):  # pragma: no cover
        return f"Boolean(value={self.value})"


TRUE = Boolean(True)
FALSE = Boolean(False)


class StringLiteral(Literal):
    __slots__ = ()

    def __init__(self, value: str):
        super().__init__(value)

    def __eq__(self, other: object):
        return isinstance(other, StringLiteral) and self.value == other.value

    def __repr__(self):  # pragma: no cover
        return f"StringLiteral(value='{self.value}')"


class IntegerLiteral(Literal):
    __slots__ = ()

    def __init__(self, value: int):
        super().__init__(value)

    def __eq__(self, other: object):
        return isinstance(other, IntegerLiteral) and self.value == other.value

    def __repr__(self):  # pragma: no cover
        return f"IntegerLiteral(value={self.value})"


class FloatLiteral(Literal):
    __slots__ = ()

    def __init__(self, value: float):
        super().__init__(value)

    def __eq__(self, other: object):
        return isinstance(other, FloatLiteral) and self.value == other.value

    def __repr__(self):  # pragma: no cover
        return f"FloatLiteral(value={self.value})"


class IdentifierPathElement(Literal):
    __slots__ = ()

    def __eq__(self, other: object):
        return isinstance(other, IdentifierPathElement) and self.value == other.value

    def __repr__(self):  # pragma: no cover
        return f"IdentifierPathElement(value={self.value})"

    def __str__(self):
        return str(self.value)


IdentifierPath = List[Union[IdentifierPathElement, "Identifier"]]


class Identifier(Expression):
    __slots__ = "path"

    def __init__(self, path: IdentifierPath):
        self.path = path

    def __eq__(self, other: object):
        return isinstance(other, Identifier) and self.path == other.path

    def __repr__(self):  # pragma: no cover
        return f"Identifier(path={self.path})"

    def __str__(self):
        return ".".join(str(elem) for elem in self.path)

    def __hash__(self):
        return hash(str(self))

    def evaluate(self, context: Context) -> object:
        path: List[Any] = [elem.evaluate(context) for elem in self.path]
        return context.get(path)


class PrefixExpression(Expression):
    __slots__ = ("operator", "right")

    def __init__(self, operator: str, right: Expression):
        self.operator = operator
        self.right = right

    def __eq__(self, other: object):
        return (
            isinstance(other, PrefixExpression)
            and self.operator == other.operator
            and self.right == other.right
        )

    def __repr__(self):  # pragma: no cover
        return f"PrefixExpression(operator='{self.operator}', right={self.right!r})"

    def __str__(self):
        return f"({self.operator}{self.right})"

    def evaluate(self, context: Context):
        right = self.right.evaluate(context)

        if self.operator == "-":
            if isinstance(right, (int, float)):
                return -right
            raise LiquidTypeError(f"unknown operator {self.operator}{self.right}")

        raise LiquidTypeError(f"unknown operator {self.operator}")


class InfixExpression(Expression):
    __slots__ = ("left", "operator", "right")

    def __init__(
        self,
        left: Expression,
        operator: str,
        right: Expression,
    ):
        self.left = left
        self.operator = operator
        self.right = right

    def __eq__(self, other: object):
        return (
            isinstance(other, InfixExpression)
            and self.left == other.left
            and self.operator == other.operator
            and self.right == other.right
        )

    def __repr__(self):  # pragma: no cover
        return (
            f"InfixExpression(left={self.left!r}, "
            f"operator='{self.operator}', right={self.right!r})"
        )

    def __str__(self):
        return f"({self.left} {self.operator} {self.right})"

    def evaluate(self, context: Context):
        left = self.left.evaluate(context)
        right = self.right.evaluate(context)
        return compare(left, self.operator, right)


class Filter:
    __slots__ = ("name", "args")

    def __init__(self, name: str, args: List[Expression]):
        self.name = name
        self.args = args

    def __eq__(self, other: object):
        return (
            isinstance(other, Filter)
            and self.name == other.name
            and self.args == other.args
        )

    def __repr__(self):  # pragma: no cover
        return f"Filter(name='{self.name}', args={self.args})"

    def __str__(self):
        args_str = ", ".join([str(arg) for arg in self.args])
        if args_str:
            return f"{self.name}: {args_str})"
        return f"{self.name}"

    def evaluate_args(self, context: Context):
        return [arg.evaluate(context) for arg in self.args]


class BooleanExpression(Expression):
    __slots__ = ("expression",)

    def __init__(self, expression: Expression):
        self.expression = expression

    def __eq__(self, other: object):
        return (
            isinstance(other, BooleanExpression) and self.expression == other.expression
        )

    def __str__(self):
        expr_str = str(self.expression)
        if not expr_str.startswith("("):
            return f"({expr_str})"
        return expr_str

    def __repr__(self):  # pragma: no cover
        return f"BooleanExpression(expression={self.expression!r})"

    def evaluate(self, context: Context) -> bool:
        return is_truthy(self.expression.evaluate(context))


class FilteredExpression(Expression):
    __slots__ = ("expression", "filters")

    def __init__(self, expression: Expression, filters: Optional[List[Filter]] = None):
        self.expression = expression
        self.filters = filters or []

    def __eq__(self, other: object):
        return (
            isinstance(other, FilteredExpression)
            and self.expression == other.expression
            and self.filters == other.filters
        )

    def __str__(self):
        filter_str = " | ".join([str(filter) for filter in self.filters])

        if filter_str:
            return f"{self.expression} | {filter_str}"
        return str(self.expression)

    def __repr__(self):  # pragma: no cover
        return (
            f"FilteredExpression(expression={self.expression!r}, "
            "filters={self.filters})"
        )

    def evaluate(self, context: Context) -> object:
        result = self.expression.evaluate(context)

        for fltr in self.filters:
            func = context.filter(fltr.name)
            try:
                result = func(result, *fltr.evaluate_args(context))
            except Exception as err:
                # Any exception causes us to abort the filter chain and discard the
                # result. Nothing will be rendered.
                if not issubclass(type(err), Error):
                    raise Error(
                        f"filter '{fltr.name}': unexpected error: {err}"
                    ) from err
                raise

        return result


class AssignmentExpression(Expression):
    __slots__ = ("name", "expression")

    def __init__(self, name: str, expression: Expression):
        self.name = name
        self.expression = expression

    def __eq__(self, other: object):
        return (
            isinstance(other, AssignmentExpression)
            and self.name == other.name
            and self.expression == other.expression
        )

    def __str__(self):
        return f"{self.name} = {self.expression}"

    def __repr__(self):  # pragma: no cover
        return (
            f"AssignmentExpression(name='{self.name}', expression={self.expression!r})"
        )

    def evaluate(self, context: Context) -> str:
        result = self.expression.evaluate(context)
        context.assign(key=self.name, val=result)
        return ""


LoopArgument = Optional[Union[IntegerLiteral, Identifier]]


class LoopExpression(Expression):
    """Loop expression tree/object model, as returned by parse_loop_expression.

    Start, stop, limit and offset can be integer literals or identifiers
    that resolve to an integer at render time.

    `cols` is used by the built-in tag `tablerow` only. It will be ignored
    by the `for` tag.
    """

    __slots__ = (
        "name",
        "identifier",
        "start",
        "stop",
        "limit",
        "offset",
        "cols",
        "reversed",
    )

    def __init__(
        self,
        name: str,
        identifier: LoopArgument = None,
        start: LoopArgument = None,
        stop: LoopArgument = None,
        limit: LoopArgument = None,
        offset: LoopArgument = None,
        cols: LoopArgument = None,
        reversed_: bool = False,
    ):
        self.name = name
        self.identifier = identifier
        self.start = start
        self.stop = stop
        self.limit = limit
        self.offset = offset
        self.cols = cols
        self.reversed = reversed_

    def __eq__(self, other: object):
        return (
            isinstance(other, LoopExpression)
            and self.name == other.name
            and self.identifier == other.identifier
            and self.start == other.start
            and self.stop == other.stop
            and self.limit == other.limit
            and self.offset == other.offset
            and self.cols == other.cols
            and self.reversed == other.reversed
        )

    def __str__(self):
        buf = [f"{self.name} in"]

        if self.identifier:
            buf.append(str(self.identifier))
        elif self.start and self.stop:
            buf.append(f"({self.start}..{self.stop})")

        if self.limit:
            buf.append(f"limit:{self.limit}")

        if self.offset:
            buf.append(f"offset:{self.offset}")

        if self.cols:
            buf.append(f"cols:{self.cols}")

        if self.reversed:
            buf.append("reversed")

        return " ".join(buf)

    def __repr__(self):  # pragma: no cover
        return (
            f"LoopExpression(name='{self.name}', identifier={self.identifier}, "
            f"start={self.start}, stop={self.stop}, limit={self.limit}, "
            f"offset={self.offset}, cols={self.cols}, reversed={self.reversed})"
        )

    def evaluate(self, context: Context) -> Tuple[Iterator[Any], int]:
        if self.identifier:
            # An identifier that must resolve to a list or dict in the current
            # global or local namespaces.
            obj = self.identifier.evaluate(context)

            if isinstance(obj, abc.Mapping):
                length = len(obj)
                loop_iter = obj.items()
            elif isinstance(obj, abc.Sequence):
                length = len(obj)
                loop_iter = iter(obj)

            else:
                raise LiquidTypeError(
                    f"expected array or hash at '{self.identifier}', found '{str(obj)}'"
                )
        else:
            assert self.start is not None
            assert self.stop is not None

            start = self.start.evaluate(context)
            stop = self.stop.evaluate(context)

            assert isinstance(start, int)
            assert isinstance(stop, int)

            stop = stop + 1

            loop_iter = range(start, stop)
            length = stop - start

        limit = self.limit
        offset = self.offset

        if limit:
            limit = limit.evaluate(context)
            assert isinstance(limit, int)
            length = min(length, limit)

        if offset:
            offset = offset.evaluate(context)
            assert isinstance(offset, int)
            length = max(length - offset, 0)

        loop_iter = islice(loop_iter, offset, limit)

        if self.reversed:
            loop_iter = reversed(list(loop_iter))

        return loop_iter, length


Number = Union[int, float]


def eval_number_expression(left: Number, operator: str, right: Number) -> bool:
    if operator == "<=":
        return left <= right
    if operator == ">=":
        return left >= right
    if operator == "<":
        return left < right
    if operator == ">":
        return left > right

    raise LiquidTypeError(f"unknown operator {left} {operator} {right}")


def eval_empty_expression(left: Any, operator: str, right: Any) -> bool:
    if not isinstance(left, Empty):
        left, right = right, left

    if operator == "==":
        return left == right
    if operator in ("!=", "<>"):
        return left != right

    raise LiquidTypeError(f"unknown operator {left} {operator} {right}")


def is_truthy(obj: Any) -> bool:
    """Return True if the given object is Liquid truthy."""
    if obj in (False, None):
        return False
    return True


def compare(left: Any, operator: str, right: Any) -> bool:
    if isinstance(left, Empty) or isinstance(right, Empty):
        return eval_empty_expression(left, operator, right)

    if operator == "and":
        return is_truthy(left) and is_truthy(right)
    if operator == "or":
        return is_truthy(left) or is_truthy(right)
    if operator == "==":
        return left == right
    if operator in ("!=", "<>"):
        return left != right

    if operator == "contains":
        if isinstance(left, str):
            return str(right) in left
        if isinstance(left, (list, dict)):
            return right in left

    # FIXME: It appears that shopify will convert any illegal comparison into something
    # falsey, at least in lax mode.

    if type(left) in (int, float) and type(right) in (int, float):
        return eval_number_expression(left, operator, right)

    if type(left) != type(right):
        raise LiquidTypeError(
            f"invalid operator for types '{str(left)} {operator} {str(right)}'"
        )

    raise LiquidTypeError(f"unknown operator: {type(left)} {operator} {type(right)}")



from liquid.filter import (
    Filter,
    one_maybe_two_args_required,
    string_required,
)

# pylint: disable=arguments-differ, too-few-public-methods


class LinkToTag(Filter):

    __slots__ = ()

    name = "link_to_tag"
    with_context = True

    def __call__(self, label, tag, *, context):
        handle = context.resolve("handle", default="")
        return (
            f'<a title="Show tag {tag}" href="/collections/{handle}/{tag}">{label}</a>'
        )


class HighlightActiveTag(Filter):

    __slots__ = ()

    name = "highlight_active_tag"

    @one_maybe_two_args_required
    @string_required
    def __call__(self, tag, css_class="active"):
        return f'<span class="{css_class}">{tag}</span>'


class LinkToAddTag(Filter):

    __slots__ = ()

    name = "link_to_add_tag"
    with_context = True

    def __call__(self, label, tag, *, context):
        handle = context.resolve("handle", default="")
        tags = "+".join(set(context.resolve("current_tags", default=[]) + [tag]))
        return (
            f'<a title="Show tag {tag}" href="/collections/{handle}/{tags}">{label}</a>'
        )


class LinkToRemoveTag(Filter):

    __slots__ = ()

    name = "link_to_remove_tag"
    with_context = True

    def __call__(self, label, tag, *, context):
        handle = context.resolve("handle", default="")
        current_tags = set(context.resolve("current_tags", default=[]))
        tags = "+".join(t for t in current_tags if t != tag)
        return (
            f'<a title="Hide tag {tag}" href="/collections/{handle}/{tags}">{label}</a>'
        )



"""Liquid expression parser test cases."""

import unittest
from typing import NamedTuple, Any

from liquid.expression import (
    FilteredExpression,
    BooleanExpression,
    AssignmentExpression,
    LoopExpression,
    IdentifierPathElement,
    Identifier,
    Boolean,
    Filter,
    StringLiteral,
    IntegerLiteral,
    FloatLiteral,
    PrefixExpression,
    InfixExpression,
)
from liquid.token import (
    Token,
    TOKEN_INITIAL,
    TOKEN_LITERAL,
    TOKEN_INTEGER,
    TOKEN_NEGATIVE,
    TOKEN_FLOAT,
    TOKEN_IDENTIFIER,
    TOKEN_STRING,
    TOKEN_TRUE,
    TOKEN_FALSE,
)
from liquid.stream import TokenStream
from liquid.lex import (
    tokenize_identifier,
    tokenize_loop_expression,
    tokenize_include_expression,
    tokenize_filtered_expression,
    tokenize_boolean_expression,
    tokenize_assignment_expression,
)

from liquid.parse import (
    parse_filtered_expression,
    parse_boolean_expression,
    parse_assignment_expression,
    parse_loop_expression,
)


class Case(NamedTuple):
    """Subtest helper"""

    description: str
    expression: str
    expect: Any


MockToken = Token(1, TOKEN_INITIAL, "MOCK")


class LiquidFilteredExpressionParserTestCase(unittest.TestCase):
    """Liquid expression parser test cases."""

    def _test(self, test_cases, lex_func, parse_func):
        """Helper method for testing lists of Cases."""

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = TokenStream(lex_func(case.expression))
                expr = parse_func(tokens)
                self.assertEqual(expr, case.expect)

    def test_parse_filtered_expression(self):
        """Test that we can parse liquid statement expressions."""
        test_cases = [
            Case(
                "string literal single quotes",
                "'foobar'",
                FilteredExpression(
                    expression=StringLiteral("foobar"),
                    filters=[],
                ),
            ),
            Case(
                "string literal double quotes",
                '"foobar"',
                FilteredExpression(
                    expression=StringLiteral("foobar"),
                    filters=[],
                ),
            ),
            Case(
                "integer literal",
                "7",
                FilteredExpression(
                    expression=IntegerLiteral(7),
                    filters=[],
                ),
            ),
            Case(
                "negative integer literal statement expression",
                "-7",
                FilteredExpression(
                    expression=PrefixExpression(
                        "-",
                        right=IntegerLiteral(7),
                    ),
                    filters=[],
                ),
            ),
            Case(
                "float literal statement expression",
                "3.14",
                FilteredExpression(
                    expression=FloatLiteral(3.14),
                    filters=[],
                ),
            ),
            Case(
                "negative float literal statement expression",
                "-3.14",
                FilteredExpression(
                    expression=PrefixExpression(
                        "-",
                        right=FloatLiteral(3.14),
                    ),
                    filters=[],
                ),
            ),
            Case(
                "single identifier statement expression",
                "collection",
                FilteredExpression(
                    expression=Identifier(
                        path=[IdentifierPathElement("collection")],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier",
                "collection.products",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key",
                'collection["products"]',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key followed by index",
                'collection["products"][0]',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement(0),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key followed by identifier",
                'collection["products"].title',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("title"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by single quoted key",
                "collection['products']",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index",
                "collection.products[0]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement(0),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index from variable",
                "collection.products[i]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            Identifier(
                                path=[IdentifierPathElement("i")],
                            ),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index from chained identifier",
                "collection.products[some.['object'].count]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            Identifier(
                                path=[
                                    IdentifierPathElement("some"),
                                    IdentifierPathElement("object"),
                                    IdentifierPathElement("count"),
                                ],
                            ),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "string literal with filter",
                "'foo' | upcase",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[Filter(name="upcase", args=[])],
                ),
            ),
            Case(
                "identifier with filter",
                "collection.title | upcase",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("title"),
                        ],
                    ),
                    filters=[Filter(name="upcase", args=[])],
                ),
            ),
            Case(
                "integer literal with filter and integer argument",
                "4 | at_least: 5",
                FilteredExpression(
                    expression=IntegerLiteral(4),
                    filters=[Filter(name="at_least", args=[IntegerLiteral(5)])],
                ),
            ),
            Case(
                "float literal with filter and float argument",
                "4.1 | divided_by: 5.2",
                FilteredExpression(
                    expression=FloatLiteral(4.1),
                    filters=[Filter(name="divided_by", args=[FloatLiteral(5.2)])],
                ),
            ),
            Case(
                "string literal with filter and string argument",
                "'foo' | append: 'bar'",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[Filter(name="append", args=[StringLiteral("bar")])],
                ),
            ),
            Case(
                "string literal with filter and identifier argument",
                "'foo' | append: collection.title",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[
                        Filter(
                            name="append",
                            args=[
                                Identifier(
                                    path=[
                                        IdentifierPathElement("collection"),
                                        IdentifierPathElement("title"),
                                    ],
                                )
                            ],
                        )
                    ],
                ),
            ),
            Case(
                "string literal with filter and multiple arguments",
                '"Liquid" | slice: 2, 5',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        )
                    ],
                ),
            ),
            Case(
                "string literal with multiple filters",
                '"Liquid" | slice: 2, 5 | upcase',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        ),
                        Filter(name="upcase", args=[]),
                    ],
                ),
            ),
            Case(
                "inconsistent whitespace",
                ' "Liquid"   |slice: 2,5',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        )
                    ],
                ),
            ),
        ]

        self._test(test_cases, tokenize_filtered_expression, parse_filtered_expression)

    def test_parse_boolean_expression(self):
        """Test that we can parse boolean expressions."""
        test_cases = [
            Case(
                "string literal double quotes",
                '"foobar"',
                BooleanExpression(
                    expression=StringLiteral("foobar"),
                ),
            ),
            Case(
                "integer literal",
                "7",
                BooleanExpression(
                    expression=IntegerLiteral(7),
                ),
            ),
            Case(
                "negative integer literal statement expression",
                "-7",
                BooleanExpression(
                    expression=PrefixExpression(
                        "-",
                        right=IntegerLiteral(7),
                    ),
                ),
            ),
            Case(
                "float literal statement expression",
                "3.14",
                BooleanExpression(
                    expression=FloatLiteral(3.14),
                ),
            ),
            Case(
                "negative float literal statement expression",
                "-3.14",
                BooleanExpression(
                    expression=PrefixExpression(
                        "-",
                        right=FloatLiteral(3.14),
                    ),
                ),
            ),
            Case(
                "single identifier statement expression",
                "collection",
                BooleanExpression(
                    expression=Identifier(
                        path=[IdentifierPathElement("collection")],
                    ),
                ),
            ),
            Case(
                "chained identifier",
                "collection.products",
                BooleanExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                ),
            ),
            Case(
                "keyword true",
                "true",
                BooleanExpression(
                    expression=Boolean(True),
                ),
            ),
            Case(
                "keyword false",
                "false",
                BooleanExpression(
                    expression=Boolean(False),
                ),
            ),
            Case(
                "boolean equality",
                "true == true",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="==",
                        right=Boolean(True),
                    ),
                ),
            ),
            Case(
                "boolean inequality",
                "true != false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="!=",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "boolean inequality alternate",
                "true <> false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="<>",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "identifier greater than literal",
                "user.age > 21",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[
                                IdentifierPathElement("user"),
                                IdentifierPathElement("age"),
                            ],
                        ),
                        operator=">",
                        right=IntegerLiteral(21),
                    ),
                ),
            ),
            Case(
                "identifier less than literal",
                "age < 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator="<",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "identifier less than or equal to literal",
                "age <= 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator="<=",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "identifier greater than or equal to literal",
                "age >= 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator=">=",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "boolean or boolean",
                "true or false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="or",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "identifier contains string",
                "product.tags contains 'sale'",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[
                                IdentifierPathElement("product"),
                                IdentifierPathElement("tags"),
                            ],
                        ),
                        operator="contains",
                        right=StringLiteral("sale"),
                    ),
                ),
            ),
        ]

        self._test(test_cases, tokenize_boolean_expression, parse_boolean_expression)

    def test_parse_boolean_expression_precedence(self):
        """Test that we get the expected precedence when parsing boolean expressions."""
        test_cases = [
            Case(
                "boolean and logic operators",
                "true and false and false or true",
                "(True and (False and (False or True)))",
            ),
            Case(
                "equality and logic operators",
                "user.name == 'bob' and user.age < 50 or collection.title == 'offers'",
                "((user.name == 'bob') and ((user.age < 50) or (collection.title == 'offers')))",
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = TokenStream(tokenize_boolean_expression(case.expression))
                expr = parse_boolean_expression(tokens)
                self.assertEqual(str(expr), case.expect)

    def test_parse_assignment_expression(self):
        """Test that we can parse assignment expressions."""
        test_cases = [
            Case(
                "assign a string literal",
                "some = 'foo'",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=StringLiteral(value="foo"),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign an integer literal",
                "some = 5",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=IntegerLiteral(value=5),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a float literal",
                "some = 5.6",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=FloatLiteral(value=5.6),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a negative integer literal",
                "some = -5.6",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=PrefixExpression(
                            operator="-",
                            right=FloatLiteral(value=5.6),
                        ),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a boolean",
                "some = true",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=Boolean(value=True),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign with a filter",
                'fruits = "apples, oranges, peaches, tomatoes" | split: ", "',
                AssignmentExpression(
                    name="fruits",
                    expression=FilteredExpression(
                        expression=StringLiteral(
                            value="apples, oranges, peaches, tomatoes",
                        ),
                        filters=[Filter(name="split", args=[StringLiteral(", ")])],
                    ),
                ),
            ),
            Case(
                "assign with multiple filters and identifier arguments",
                "everything = fruits | concat: vegetables | concat: furniture",
                AssignmentExpression(
                    name="everything",
                    expression=FilteredExpression(
                        expression=Identifier(
                            path=[IdentifierPathElement("fruits")],
                        ),
                        filters=[
                            Filter(
                                name="concat",
                                args=[
                                    Identifier(
                                        path=[IdentifierPathElement("vegetables")],
                                    )
                                ],
                            ),
                            Filter(
                                name="concat",
                                args=[
                                    Identifier(
                                        path=[IdentifierPathElement("furniture")],
                                    )
                                ],
                            ),
                        ],
                    ),
                ),
            ),
        ]

        self._test(
            test_cases, tokenize_assignment_expression, parse_assignment_expression
        )

    def test_parse_loop_expression(self):
        """Test that we can parse loop expressions."""
        test_cases = [
            Case(
                "item in array",
                "product in collection.products",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in array with integer literal limit",
                "product in collection.products limit:5",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                ),
            ),
            Case(
                "item in array with identifier limit",
                "product in collection.products limit:max",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=Identifier(path=[IdentifierPathElement("max")]),
                ),
            ),
            Case(
                "item in array with integer literal offset",
                "product in collection.products offset:2",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit",
                "product in collection.products limit:5 offset:2",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit out of order",
                "product in collection.products offset:2 limit:5",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array reversed",
                "product in collection.products reversed",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    reversed_=True,
                ),
            ),
            Case(
                "item in array with all options",
                "product in collection.products offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
            Case(
                "item in range with integer literal start and stop",
                "product in (1..10)",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=IntegerLiteral(10),
                ),
            ),
            Case(
                "item in range with integer literal start and identifier stop",
                "product in (1..collection.products.size)",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("size"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in range with integer identifier start and stop",
                "product in (num..collection.products.size)",
                LoopExpression(
                    name="product",
                    start=Identifier(path=[IdentifierPathElement("num")]),
                    stop=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("size"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in range with options",
                "product in (1..10) offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=IntegerLiteral(10),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
        ]

        self._test(test_cases, tokenize_loop_expression, parse_loop_expression)



"""Test math filter functions."""
import unittest
from inspect import isclass
from typing import NamedTuple, Any, List, Dict

from liquid.environment import Environment
from liquid.exceptions import FilterArgumentError
from liquid.builtin.filters.math import (
    Abs,
    AtMost,
    AtLeast,
    Ceil,
    DividedBy,
    Floor,
    Minus,
    Plus,
    Round,
    Times,
    Modulo,
)


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class MathFilterTestCase(unittest.TestCase):
    """Test math filter functions."""

    def _test(self, filter_cls, test_cases):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        func = filter_cls(env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, FilterArgumentError
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_abs(self):
        """Test abs filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string integer",
                val="5",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative string integer",
                val="-5",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=5.1,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=5.1,
            ),
            Case(
                description="unexpected argument",
                val=-3,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Abs, test_cases)

    def test_at_most(self):
        """Test at_most filter function."""

        test_cases = [
            Case(
                description="positive integer < arg",
                val=5,
                args=[8],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive integer > arg",
                val=8,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer < arg",
                val=-8,
                args=[5],
                kwargs={},
                expect=-8,
            ),
            Case(
                description="positive integer == arg",
                val=5,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float < arg",
                val=5.4,
                args=[8.9],
                kwargs={},
                expect=5.4,
            ),
            Case(
                description="positive float > arg",
                val=8.4,
                args=[5.9],
                kwargs={},
                expect=5.9,
            ),
            Case(
                description="positive string > arg",
                val="9",
                args=["8"],
                kwargs={},
                expect=8,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(AtMost, test_cases)

    def test_at_least(self):
        """Test at_least filter function."""

        test_cases = [
            Case(
                description="positive integer < arg",
                val=5,
                args=[8],
                kwargs={},
                expect=8,
            ),
            Case(
                description="positive integer > arg",
                val=8,
                args=[5],
                kwargs={},
                expect=8,
            ),
            Case(
                description="negative integer < arg",
                val=-8,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive integer == arg",
                val=5,
                args=[5],
                kwargs={},
                expect=5,
            ),
            Case(
                description="positive float < arg",
                val=5.4,
                args=[8.9],
                kwargs={},
                expect=8.9,
            ),
            Case(
                description="positive float > arg",
                val=8.4,
                args=[5.9],
                kwargs={},
                expect=8.4,
            ),
            Case(
                description="positive string > arg",
                val="9",
                args=["8"],
                kwargs={},
                expect=9,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(AtLeast, test_cases)

    def test_ceil(self):
        """Test ceil filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="unexpected argument",
                val=-3.1,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Ceil, test_cases)

    def test_floor(self):
        """Test floor filter function."""

        test_cases = [
            Case(
                description="positive integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative integer",
                val=-5,
                args=[],
                kwargs={},
                expect=-5,
            ),
            Case(
                description="positive float",
                val=5.4,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative float",
                val=-5.4,
                args=[],
                kwargs={},
                expect=-6,
            ),
            Case(
                description="zero",
                val=0,
                args=[],
                kwargs={},
                expect=0,
            ),
            Case(
                description="positive string float",
                val="5.1",
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="negative string float",
                val="-5.1",
                args=[],
                kwargs={},
                expect=-6,
            ),
            Case(
                description="unexpected argument",
                val=-3.1,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="string not a number",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Floor, test_cases)

    def test_divided_by(self):
        """Test divided_by filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=5,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=5.0,
            ),
            Case(
                description="integer division",
                val=9,
                args=[2],
                kwargs={},
                expect=4,
            ),
            Case(
                description="float division",
                val=20,
                args=[7.0],
                kwargs={},
                expect=2.857142857142857,
            ),
            Case(
                description="string value and argument",
                val="10",
                args=["2.0"],
                kwargs={},
                expect=5.0,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(DividedBy, test_cases)

    def test_minus(self):
        """Test minus filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=8,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=8.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[2.2],
                kwargs={},
                expect=7.9,
            ),
            Case(
                description="string value and string arg",
                val="10.1",
                args=["2.2"],
                kwargs={},
                expect=7.9,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Minus, test_cases)

    def test_plus(self):
        """Test plus filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=12,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=12.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[2.2],
                kwargs={},
                expect=12.3,
            ),
            Case(
                description="string value and string arg",
                val="10.1",
                args=["2.2"],
                kwargs={},
                expect=12.3,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Plus, test_cases)

    def test_round(self):
        """Test round filter function."""

        test_cases = [
            Case(
                description="float round down",
                val=5.1,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="float round up",
                val=5.6,
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="float as a string",
                val="5.6",
                args=[],
                kwargs={},
                expect=6,
            ),
            Case(
                description="string argument",
                val=5.666,
                args=["1"],
                kwargs={},
                expect=5.7,
            ),
            Case(
                description="decimal places",
                val="5.666666",
                args=[2],
                kwargs={},
                expect=5.67,
            ),
            Case(
                description="integer",
                val=5,
                args=[],
                kwargs={},
                expect=5,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Round, test_cases)

    def test_times(self):
        """Test times filter function."""

        test_cases = [
            Case(
                description="int times int",
                val=5,
                args=[2],
                kwargs={},
                expect=10,
            ),
            Case(
                description="int times float",
                val=5,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="int times float",
                val=5,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="float times float",
                val=5.0,
                args=[2.1],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="string times string",
                val="5.0",
                args=["2.1"],
                kwargs={},
                expect=10.5,
            ),
            Case(
                description="negative multiplication",
                val=-5,
                args=[2],
                kwargs={},
                expect=-10,
            ),
            Case(
                description="missing arg",
                val=5,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, 2],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Times, test_cases)

    def test_modulo(self):
        """Test modulo filter function."""

        test_cases = [
            Case(
                description="integer value and integer arg",
                val=10,
                args=[2],
                kwargs={},
                expect=0,
            ),
            Case(
                description="integer value and float arg",
                val=10,
                args=[2.0],
                kwargs={},
                expect=0.0,
            ),
            Case(
                description="float value and float arg",
                val=10.1,
                args=[7.0],
                kwargs={},
                expect=3.1,
            ),
            Case(
                description="string value and argument",
                val="10",
                args=["2.0"],
                kwargs={},
                expect=0.0,
            ),
            Case(
                description="string not a number",
                val="foo",
                args=["2.0"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="arg string not a number",
                val="10",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many args",
                val=5,
                args=[1, "5"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string, int or float",
                val=object(),
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Modulo, test_cases)



from __future__ import annotations

from abc import ABC, abstractmethod

from typing import Protocol
from typing import Union
from typing import Type
from typing import Optional
from typing import Any

from liquid.ast import Node
from liquid.ast import IllegalNode
from liquid.exceptions import Error
from liquid.parse import eat_block
from liquid.stream import TokenStream


class Env(Protocol):
    def error(
        self: Any,
        exc: Union[Type[Error], Error],
        msg: Optional[str] = ...,
        linenum: Optional[int] = ...,
    ) -> None:
        ...


class Tag(ABC):
    """Base class for all built-in and custom template tags."""

    block = True
    name = ""

    def __init__(self, env: Env):
        self.env = env

    def get_node(self, stream: TokenStream) -> Node:
        """Wraps `Tag.parse`, possibly returning an `IllegalNode`."""
        tok = stream.current
        try:
            return self.parse(stream)
        except Error as err:
            if not err.linenum:
                err.linenum = tok.linenum

            self.env.error(err)

            if self.block and hasattr(self, "end"):
                eat_block(stream, (getattr(self, "end"),))

            return IllegalNode(tok)

    @abstractmethod
    def parse(self, stream: TokenStream) -> Node:
        """Return a parse tree node py parsing tokens from the given stream."""



"""Liquid template parser."""

from __future__ import annotations

from enum import IntEnum, auto
from functools import lru_cache

from typing import Tuple
from typing import List
from typing import Union
from typing import Optional
from typing import NamedTuple
from typing import Dict
from typing import Protocol
from typing import Any

from liquid.exceptions import LiquidSyntaxError, Error
from liquid.mode import Mode, error

from liquid import expression
from liquid import ast
from liquid.stream import TokenStream

from liquid.token import TOKEN_ILLEGAL
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_STATEMENT
from liquid.token import TOKEN_LITERAL
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NEGATIVE
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_CONTAINS
from liquid.token import TOKEN_IN
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_LIMIT
from liquid.token import TOKEN_OFFSET
from liquid.token import TOKEN_REVERSED
from liquid.token import TOKEN_COLS
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_LBRACKET
from liquid.token import TOKEN_RBRACKET
from liquid.token import TOKEN_ASSIGN
from liquid.token import TOKEN_AND
from liquid.token import TOKEN_OR
from liquid.token import TOKEN_EQ
from liquid.token import TOKEN_NE
from liquid.token import TOKEN_LG
from liquid.token import TOKEN_LT
from liquid.token import TOKEN_GT
from liquid.token import TOKEN_LE
from liquid.token import TOKEN_GE
from liquid.token import reverse_operators
from liquid.token import Token


class Tag(Protocol):

    block: bool

    def get_node(self: Any, stream: TokenStream) -> ast.Node:
        ...


class Env(Protocol):

    tags: Dict[str, Tag]
    mode: Mode


class Parser:
    __slots__ = ("tags", "mode", "illegal", "literal", "statement")

    def __init__(self, env: Env):
        self.tags = env.tags
        self.mode = env.mode

        self.illegal = self.tags[TOKEN_ILLEGAL]
        self.literal = self.tags[TOKEN_LITERAL]
        self.statement = self.tags[TOKEN_STATEMENT]

    def parse(self, stream: TokenStream) -> ast.ParseTree:
        root = ast.ParseTree()

        while stream.current.type != TOKEN_EOF:
            try:
                if node := self.parse_statement(stream):
                    root.statements.append(node)
            except Error as err:
                error(self.mode, err, linenum=stream.current.linenum)

            stream.next_token()

        return root

    def parse_statement(self, stream: TokenStream) -> ast.Node:
        """"""
        if stream.current.test(TOKEN_STATEMENT):
            node = self.statement.get_node(stream)
        elif stream.current.test(TOKEN_TAG):
            tag = self.tags.get(stream.current.value, self.illegal)

            if tag.block:
                stream.balancing_stack.append(tag)

            node = tag.get_node(stream)

            # Tag parse functions can choose to return an IllegalNode.
            if node.__class__ == "IllegalNode":
                raise LiquidSyntaxError(
                    f"unexpected tag '{node.token().value}'",
                    linenum=node.token().linenum,
                )

            if tag.block:
                popped = stream.balancing_stack.pop()
                if stream.current.value != popped.end:
                    raise LiquidSyntaxError(
                        f"expected {popped.end}, found {stream.current.value}",
                        linenum=stream.current.linenum,
                    )
        else:
            expect(stream, TOKEN_LITERAL)
            node = self.literal.get_node(stream)

        return node

    def parse_block(self, stream: TokenStream, end: Tuple[str, ...]) -> ast.BlockNode:
        block = ast.BlockNode(stream.current)

        while stream.current.type != TOKEN_EOF:
            if stream.current.type == TOKEN_TAG and stream.current.value in end:
                break
            stmt = self.parse_statement(stream)
            if stmt:
                block.statements.append(stmt)
            stream.next_token()

        return block


def _expect(tok: Token, typ: str, value: Optional[str] = None) -> None:
    if tok.type != typ or (value is not None and tok.value != value):
        _typ = reverse_operators.get(tok.type, tok.type)
        _expected_typ = reverse_operators.get(typ, typ)
        if value is not None:
            msg = (
                f"expected {_expected_typ} with value '{value}', "
                f"found {_typ} with value '{tok.value}'"
            )
        else:
            msg = f"expected '{_expected_typ}', found '{_typ}'"
        raise LiquidSyntaxError(msg, linenum=tok.linenum)


def expect(stream: TokenStream, typ: str, value: Optional[str] = None) -> None:
    _expect(stream.current, typ, value)


def expect_peek(stream: TokenStream, typ: str, value: Optional[str] = None) -> None:
    _expect(stream.peek, typ, value)


def is_end_tag(tok: Token) -> bool:
    return tok.value.startswith("end")


def eat_block(stream: TokenStream, end: Tuple[str, ...]) -> None:
    """Advance the stream pointer past the next end tag.

    This is used to discard blocks after a syntax error is found, in the hope
    that we can continue to parse more of the stream after the offending block.
    """
    while stream.current.type != TOKEN_EOF:
        if stream.current.type == TOKEN_TAG and stream.current.value in end:
            break
        stream.next_token()


class Precedence(IntEnum):
    """Operator precedence."""

    LOWEST = 1
    LOGICALRIGHT = 49
    LOGICAL = 50
    RELATIONAL = auto()
    MEMBERSHIP = auto()
    PREFIX = auto()


PRECEDENCES = {
    TOKEN_IDENTIFIER: Precedence.LOWEST,
    TOKEN_EQ: Precedence.RELATIONAL,
    TOKEN_LT: Precedence.RELATIONAL,
    TOKEN_GT: Precedence.RELATIONAL,
    TOKEN_NE: Precedence.RELATIONAL,
    TOKEN_LG: Precedence.RELATIONAL,
    TOKEN_LE: Precedence.RELATIONAL,
    TOKEN_GE: Precedence.RELATIONAL,
    TOKEN_CONTAINS: Precedence.MEMBERSHIP,
    TOKEN_AND: Precedence.LOGICAL,
    TOKEN_OR: Precedence.LOGICAL,
}

RIGHT_ASSOCIATIVE = frozenset((TOKEN_AND, TOKEN_OR))

IDENTIFIER_TOKENS = frozenset(
    (TOKEN_IDENTIFIER, TOKEN_INTEGER, TOKEN_DOT, TOKEN_LBRACKET, TOKEN_STRING)
)


def peek_precedence(stream: TokenStream) -> Precedence:
    """Return the precedence of the next token in the stream."""
    try:
        precedence = PRECEDENCES[stream.peek.type]
    except KeyError as err:
        raise LiquidSyntaxError(f"unknown operator '{stream.peek.value}'") from err
    return precedence


def current_precedence(stream: TokenStream) -> Precedence:
    """Return the precedence of the current token in the stream."""
    precedence = PRECEDENCES.get(stream.current.type, Precedence.LOWEST)
    if stream.current.type in RIGHT_ASSOCIATIVE:
        precedence -= 1
    return Precedence(precedence)


def parse_filters(stream: TokenStream) -> List[expression.Filter]:
    """Keep reading filters from the token stream until end of expression."""
    filters = []
    while stream.current.type != TOKEN_EOF:
        filters.append(parse_filter(stream))
    return filters


def parse_filter(stream: TokenStream) -> expression.Filter:
    """Read a single filter with optional arguments from the token stream."""
    # Eat the pipe token
    expect(stream, TOKEN_PIPE)
    stream.next_token()

    # Read the filter's name
    expect(stream, TOKEN_IDENTIFIER)
    filter_name = stream.current.value
    stream.next_token()

    #
    args = []
    if stream.current.type == TOKEN_COLON:
        stream.next_token()

        while stream.current.type not in (TOKEN_PIPE, TOKEN_EOF):
            args.append(parse_expression(stream))
            stream.next_token()

            if stream.current.type == TOKEN_COMMA:
                stream.next_token()

    return expression.Filter(filter_name, args)


def parse_boolean(stream: TokenStream) -> expression.Boolean:
    """Read a boolean literal (true or false) from the token stream."""
    if stream.current.type == TOKEN_TRUE:
        return expression.TRUE
    return expression.FALSE


def parse_nil(_: TokenStream) -> expression.Nil:
    """Read a 'nil' keyword from the token stream."""
    return expression.NIL


def parse_empty(_: TokenStream) -> expression.Empty:
    """Read a 'empty' keyword from the token stream."""
    return expression.EMPTY


def parse_string_literal(stream: TokenStream) -> expression.StringLiteral:
    """Read a string from the token stream."""
    return expression.StringLiteral(value=stream.current.value)


def parse_integer_literal(stream: TokenStream) -> expression.IntegerLiteral:
    """Read an integer from the token stream."""
    return expression.IntegerLiteral(value=int(stream.current.value))


def parse_float_literal(stream: TokenStream) -> expression.FloatLiteral:
    """Read a float from the token stream."""
    return expression.FloatLiteral(value=float(stream.current.value))


def parse_identifier(stream: TokenStream) -> expression.Identifier:
    """Read an identifier from the token stream.

    <ident>.<ident>
    <ident>["<ident>"]
    <ident>["<ident>"].<ident>
    <ident>[<ident --> int/str>]
    <ident>[<ident>.<ident --> int/str>]
    <ident>[<int>]
    <ident>[<int>].<ident>
    """
    path = []

    while stream.current.type in IDENTIFIER_TOKENS:
        if stream.current.type == TOKEN_IDENTIFIER:
            path.append(expression.IdentifierPathElement(stream.current.value))

        elif stream.current.type == TOKEN_INTEGER:
            path.append(expression.IdentifierPathElement(int(stream.current.value)))

        elif stream.current.type == TOKEN_LBRACKET:
            stream.next_token()  # Eat open bracket

            if stream.current.type == TOKEN_STRING:
                path.append(expression.IdentifierPathElement(stream.current.value))

            elif stream.current.type == TOKEN_INTEGER:
                path.append(expression.IdentifierPathElement(int(stream.current.value)))

            elif stream.current.type == TOKEN_IDENTIFIER:
                # Recursive call to parse_identifier. If it's not a string or
                # integer, anything inside a pair of square brackets could be
                # another identifier that resolves to a string or integer.
                path.append(parse_identifier(stream))
            else:
                raise LiquidSyntaxError(
                    f"invalid identifier, found {stream.current.type}"
                )

            expect_peek(stream, TOKEN_RBRACKET)
            stream.next_token()  # Eat close bracket

        elif stream.current.type == TOKEN_DOT:
            pass
        else:
            raise LiquidSyntaxError(f"invalid identifier, found {stream.current.type}")

        stream.next_token()

    stream.push(stream.current)
    return expression.Identifier(path)


class RangeOption(NamedTuple):
    tok: Token
    arg: Optional[Union[expression.Identifier, expression.IntegerLiteral]]


def parse_range_argument(
    stream: TokenStream,
) -> Union[expression.Identifier, expression.IntegerLiteral]:
    if stream.current.type == TOKEN_IDENTIFIER:
        arg = parse_identifier(stream)
    elif stream.current.type == TOKEN_INTEGER:
        arg = parse_integer_literal(stream)
    else:
        raise LiquidSyntaxError(
            "invalid range expression, expected an integer, "
            f"found a {stream.current.type}"
        )
    return arg


def parse_string_or_identifier(
    stream: TokenStream,
    linenum: Optional[int] = None,
) -> ast.Expression:
    if stream.current.type == TOKEN_IDENTIFIER:
        expr = parse_identifier(stream)
    elif stream.current.type == TOKEN_STRING:
        expr = parse_string_literal(stream)
    else:
        _typ = reverse_operators.get(stream.current.type, stream.current.type)
        msg = f"expected {TOKEN_IDENTIFIER} or {TOKEN_STRING}, found {_typ}"
        raise LiquidSyntaxError(msg, linenum=linenum or stream.current.linenum)

    return expr


def parse_unchained_identifier(
    stream: TokenStream,
    linenum: Optional[int] = None,
) -> expression.Identifier:
    expect(stream, TOKEN_IDENTIFIER)
    tok = stream.current
    ident = parse_identifier(stream)

    if len(ident.path) != 1:
        raise LiquidSyntaxError(
            f"invalid identifier '{ident}'", linenum=linenum or tok.linenum
        )

    return ident


def parse_range_option(stream: TokenStream) -> RangeOption:
    tok = stream.current
    if stream.current.type in (TOKEN_LIMIT, TOKEN_OFFSET, TOKEN_COLS):
        expect_peek(stream, TOKEN_COLON)
        stream.next_token()
        stream.next_token()  # Eat COLON
        arg = parse_range_argument(stream)
        stream.next_token()
        opt = RangeOption(tok=tok, arg=arg)
    else:
        expect(stream, TOKEN_REVERSED)
        stream.next_token()  # Eat REVERSED
        opt = RangeOption(tok=tok, arg=None)

    return opt


def parse_prefix_expression(stream: TokenStream) -> expression.PrefixExpression:
    tok = stream.current
    stream.next_token()

    exp = expression.PrefixExpression(
        tok.value,
        right=parse_expression(stream, precedence=Precedence.PREFIX),
    )

    return exp


def parse_infix_expression(
    stream: TokenStream, left: expression.Expression
) -> expression.InfixExpression:
    tok = stream.current
    precedence = current_precedence(stream)
    stream.next_token()

    exp = expression.InfixExpression(
        left=left,
        operator=tok.value,
        right=parse_expression(stream, precedence),
    )
    return exp


prefix_funcs = {
    TOKEN_FALSE: parse_boolean,
    TOKEN_TRUE: parse_boolean,
    TOKEN_NIL: parse_nil,
    TOKEN_EMPTY: parse_empty,
    TOKEN_NEGATIVE: parse_prefix_expression,
    TOKEN_STRING: parse_string_literal,
    TOKEN_INTEGER: parse_integer_literal,
    TOKEN_FLOAT: parse_float_literal,
    TOKEN_IDENTIFIER: parse_identifier,
}

infix_funcs = {
    TOKEN_EQ: parse_infix_expression,
    TOKEN_OR: parse_infix_expression,
    TOKEN_AND: parse_infix_expression,
    TOKEN_LT: parse_infix_expression,
    TOKEN_GT: parse_infix_expression,
    TOKEN_NE: parse_infix_expression,
    TOKEN_LG: parse_infix_expression,
    TOKEN_LE: parse_infix_expression,
    TOKEN_GE: parse_infix_expression,
    TOKEN_CONTAINS: parse_infix_expression,
}


def parse_expression(
    stream: TokenStream,
    precedence: Precedence = Precedence.LOWEST,
) -> expression.Expression:
    """Parse a literal or logical expression."""
    prefix = prefix_funcs.get(stream.current.type)
    if prefix is None:
        operator = reverse_operators.get(stream.current.type, stream.current.value)
        raise LiquidSyntaxError(f"unknown prefix operator '{operator}'")

    left = prefix(stream)

    while (
        stream.peek.type != TOKEN_EOF
        and stream.peek.type not in (TOKEN_PIPE, TOKEN_COMMA)
        and precedence < peek_precedence(stream)
    ):
        infix = infix_funcs.get(stream.peek.type)
        if infix is None:
            return left

        stream.next_token()
        left = infix(stream, left)

    return left


def parse_filtered_expression(stream: TokenStream) -> expression.FilteredExpression:
    """Parse a statement (output) expression with optional filters.

    A statement expression is assumed to start with exactly one identifier, followed by
    zero or more filters.
    """
    expr = parse_expression(stream)
    stream.next_token()
    filters = parse_filters(stream)
    return expression.FilteredExpression(expr, filters)


def parse_boolean_expression(stream: TokenStream) -> expression.Expression:
    """Parse a liquid expression that evaluates to a Boolean value.

    This is primarily used by control flow tags like `if`, `unless` and `case`/`when`.

    Logical operators (`and` and `or`) are right associative. They are evaluated from
    right to left. The use of parentheses to control operator precedence is not allowed.
    """
    if stream.current.type == TOKEN_EOF:
        # Empty expression.
        return expression.Nil()
    return expression.BooleanExpression(expression=parse_expression(stream))


def parse_assignment_expression(stream: TokenStream) -> expression.AssignmentExpression:
    """Parse a liquid assignment expression, as one might find in an `assign` tag.

    This is essentially the same as a parse_filtered_expression, but with
    an additional name to bind the expression result to.
    """
    expect(stream, TOKEN_IDENTIFIER)
    expect_peek(stream, TOKEN_ASSIGN)

    name = parse_unchained_identifier(stream)

    stream.next_token()
    stream.next_token()  # Eat ASSIGN TOKEN

    expr = parse_filtered_expression(stream)

    return expression.AssignmentExpression(name=str(name), expression=expr)


def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
    """Parse a liquid loop expression, as one might find in a `for` tag.

    If any of the optional parameters are duplicated, the last (right most) is
    used.
    """
    expect(stream, TOKEN_IDENTIFIER)
    expect_peek(stream, TOKEN_IN)

    expr = expression.LoopExpression(name=stream.current.value)

    stream.next_token()
    stream.next_token()  # Eat TOKEN_IN

    if stream.current.type == TOKEN_IDENTIFIER:
        # Identifier should resolve to an array or hash/map at render time.
        expr.identifier = parse_identifier(stream)
        stream.next_token()
    elif stream.current.type == TOKEN_LPAREN:
        # Start of a range expression (<int or id>..<int or id>)
        stream.next_token()
        start = parse_range_argument(stream)

        # assert stream.peek.type == TOKEN_RANGE
        expect_peek(stream, TOKEN_RANGE)

        stream.next_token()
        stream.next_token()  # Eat TOKEN_RANGE
        stop = parse_range_argument(stream)

        # assert stream.peek.type == TOKEN_RPAREN
        expect_peek(stream, TOKEN_RPAREN)

        expr.start = start
        expr.stop = stop

        stream.next_token()
        stream.next_token()  # Eat TOKEN_RPAREN
    else:
        raise LiquidSyntaxError("invalid range expression")

    # Optional range modifiers can appear in any order.
    while stream.current.type in (
        TOKEN_LIMIT,
        TOKEN_OFFSET,
        TOKEN_COLS,
        TOKEN_REVERSED,
    ):
        opt = parse_range_option(stream)

        if opt.tok.type == TOKEN_REVERSED:
            expr.reversed = True
        elif opt.tok.type == TOKEN_LIMIT:
            expr.limit = opt.arg
        elif opt.tok.type == TOKEN_OFFSET:
            expr.offset = opt.arg
        elif opt.tok.type == TOKEN_COLS:
            expr.cols = opt.arg

    return expr


@lru_cache
def get_parser(env: Env) -> Parser:
    return Parser(env)



"""Liquid template and expression parsing tests."""

import unittest
from typing import NamedTuple

from liquid.environment import Environment
from liquid.mode import Mode

from liquid.builtin import (
    literal,
    statement,
    if_tag,
    comment_tag,
    unless_tag,
    case_tag,
    for_tag,
    tablerow_tag,
    capture_tag,
    cycle_tag,
    assign_tag,
    increment_tag,
    decrement_tag,
    echo_tag,
    include_tag,
    render_tag,
)


class Case(NamedTuple):
    description: str
    template: str
    expect: str
    expect_statements: int = 1


class ParserTestCase(unittest.TestCase):
    def _test(self, test_cases, instance, mode: Mode = Mode.STRICT):
        """Helper method for testing lists of `Case`s."""
        env = Environment()
        env.mode = mode

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                self.assertIsNotNone(template)
                self.assertEqual(len(template.tree.statements), case.expect_statements)
                if case.expect_statements:
                    self.assertIsInstance(template.tree.statements[0], instance)
                    self.assertEqual(str(template.tree), case.expect)

    def test_template_literal(self):
        """Test that we can parse a template literal."""
        test_cases = [
            Case("plain html", "<HTML>Some text</HTML>", "<HTML>Some text</HTML>"),
        ]

        self._test(test_cases, literal.LiteralNode)

    def test_output_statement(self):
        """Test that we can parse output statements."""
        test_cases = [
            Case("string literal", r"{{ 'hello' }}", "`'hello'`"),
            Case("identifier", r"{{ product }}", "`product`"),
            Case(
                "identifier with a filter",
                r"{{ product.title | upcase }}",
                "`product.title | upcase`",
            ),
        ]

        self._test(test_cases, statement.StatementNode)

    def test_illegal_tag(self):
        """Test that we can handle illegal tokens in lax mode."""

        test_cases = [
            Case(
                "unknown tag name",
                "{% foobar some %}{% endfoobar %}",
                "",
                expect_statements=1,
            ),
            Case(
                "unknown tag name with surrounding literal",
                "<HTML>{% foobar %}{% endfoobar %}</HTML>",
                "<HTML></HTML>",
                expect_statements=3,
            ),
        ]

        env = Environment(tolerance=Mode.LAX)

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                self.assertIsNotNone(template)
                self.assertEqual(len(template.tree.statements), case.expect_statements)

                if case.expect_statements:
                    self.assertEqual(str(template.tree), case.expect)

    def test_if_tag(self):
        """Test that we can parse an if expression."""
        test_cases = [
            Case(
                "condition with literal consequence",
                "{% if product == 'foo' %}foo{% endif %}",
                "if (product == 'foo') { foo }",
            ),
            Case(
                "condition with literal consequence and following literal",
                "{% if product == 'foo' %}foo{% endif %} hello",
                "if (product == 'foo') { foo } hello",
                expect_statements=2,
            ),
            Case(
                "condition with literal consequence and literal alternative",
                "{% if product == 'foo' %}foo{% else %}bar{% endif %}",
                "if (product == 'foo') { foo } else { bar }",
            ),
            Case(
                "condition with literal and object consequence",
                "{% if product == 'foo' %}foo {{collection}} bar{% endif %}",
                "if (product == 'foo') { foo `collection` bar }",
            ),
            Case(
                "condition with nested condition in the consequence block",
                "{% if product == 'foo' %}{% if title == 'baz' %}baz{% endif %}{% endif %}",
                "if (product == 'foo') { if (title == 'baz') { baz } }",
            ),
            Case(
                "condition with nested, alternative condition in the consequence block",
                "{% if product == 'foo' %}{% if title == 'baz' %}baz{% else %}bar{% endif %}{% endif %}",
                "if (product == 'foo') { if (title == 'baz') { baz } else { bar } }",
            ),
            Case(
                "condition with nested condition in the alternative block",
                "{% if product == 'foo' %}foo{% else %}{% if title == 'baz' %}bar{% endif %}{% endif %}",
                "if (product == 'foo') { foo } else { if (title == 'baz') { bar } }",
            ),
            Case(
                "condition with conditional alternative",
                "{% if product == 'foo' %}foo{% elsif product == 'bar' %}bar{% endif %}",
                "if (product == 'foo') { foo } elsif (product == 'bar') { bar }",
            ),
            Case(
                "condition with conditional and final alternatives",
                "{% if product == 'foo' %}foo{% elsif product == 'bar' %}bar{% else %}baz{% endif %}",
                "if (product == 'foo') { foo } elsif (product == 'bar') { bar } else { baz }",
            ),
            Case(
                "condition with multiple alternatives",
                "{% if product == 'foo' %}foo{% elsif product == 'bar' %}bar{% elsif x == 1 %}baz{% endif %}",
                "if (product == 'foo') { foo } elsif (product == 'bar') { bar } elsif (x == 1) { baz }",
            ),
        ]

        self._test(test_cases, if_tag.IfNode)

    def test_comment_tag(self):
        """Test that we can parse comment tags."""
        test_cases = [
            Case("simple comment", "{% comment %}foo{% endcomment %}", "/* */"),
            Case(
                "commented object",
                "{% comment %}{{ product }}{% endcomment %}",
                "/* */",
            ),
        ]

        self._test(test_cases, comment_tag.CommentNode)

    def test_unless_tag(self):
        """Test that we can parse `unless` tags."""
        test_cases = [
            Case(
                "simple unless",
                "{% unless x == 1 %}foo{% endunless %}",
                "if !(x == 1) { foo }",
            ),
            Case(
                "unless with object and literal",
                "{% unless x == 1 %}foo {{ title }} bar{% endunless %}",
                "if !(x == 1) { foo `title` bar }",
            ),
        ]

        self._test(test_cases, unless_tag.UnlessNode)

    def test_case_tag(self):
        """Test that we can parse `case`/`when` tags."""
        test_cases = [
            Case(
                "simple case",
                "{% case title %}{% when 'foo' %}foo{% when 'bar' %}bar{% endcase %}",
                "if (title == 'foo') { foo } elsif (title == 'bar') { bar }",
            ),
            Case(
                "case with default",
                "{% case title %}{% when 'foo' %}foo{% else %}bar{% endcase %}",
                "if (title == 'foo') { foo } else { bar }",
            ),
            Case(
                "case with no whens",
                "{% case title %}{% else %}bar{% endcase %}",
                "if (False) { } else { bar }",
            ),
            Case(
                "case with no whens or default",
                "{% case title %}{% endcase %}",
                "if (False) { }",
            ),
            Case(
                "simple case with whitespace",
                "{% case title %} {% when 'foo' %}foo{% when 'bar' %}bar{% endcase %}",
                "if (title == 'foo') { foo } elsif (title == 'bar') { bar }",
            ),
        ]

        self._test(test_cases, case_tag.CaseNode)

    def test_for_tag(self):
        """Test that we can parse for loop tags."""
        test_cases = [
            Case(
                "simple for loop",
                "{% for prod in collection %}{{ prod.title }}{% endfor %}",
                "for (prod in collection) { `prod.title` }",
            ),
            Case(
                "for loop with default",
                "{% for prod in collection %}{{ prod.title }}{% else %}foo{% endfor %}",
                "for (prod in collection) { `prod.title` } else { foo }",
            ),
            Case(
                "for loop with break",
                "{% for prod in collection %}{% if prod.title %}{{ prod.title }}{% else %}{% break %}{% endif %}{% endfor %}",
                "for (prod in collection) { if (prod.title) { `prod.title` } else { `break` } }",
            ),
            Case(
                "for loop with continue",
                "{% for prod in collection %}{% if prod.title %}{{ prod.title }}{% else %}{% continue %}{% endif %}{% endfor %}",
                "for (prod in collection) { if (prod.title) { `prod.title` } else { `continue` } }",
            ),
            Case(
                "for loop with limit",
                "{% for prod in collection limit:2 %}{% if prod.title %}{{ prod.title }}{% else %}{% continue %}{% endif %}{% endfor %}",
                "for (prod in collection limit:2) { if (prod.title) { `prod.title` } else { `continue` } }",
            ),
            Case(
                "for loop with offset",
                "{% for prod in collection offset:1 %}{% if prod.title %}{{ prod.title }}{% else %}{% continue %}{% endif %}{% endfor %}",
                "for (prod in collection offset:1) { if (prod.title) { `prod.title` } else { `continue` } }",
            ),
        ]

        self._test(test_cases, for_tag.ForNode)

    def test_tablerow_tag(self):
        """Test that we can parse tablerow tags."""
        test_cases = [
            Case(
                "simple table row",
                "{% tablerow prod in collection.products %}{{ prod.title }}{% endtablerow %}",
                "tablerow(prod in collection.products) { `prod.title` }",
            ),
            Case(
                "table row with whitespace",
                "{% tablerow prod in collection.products %} {{ prod.title }} {% endtablerow %}",
                "tablerow(prod in collection.products) {  `prod.title`  }",
            ),
            Case(
                "table row with arguments",
                "{% tablerow prod in collection.products cols:2 limit:3 %}{{ prod.title }}{% endtablerow %}",
                "tablerow(prod in collection.products limit:3 cols:2) { `prod.title` }",
            ),
        ]

        self._test(test_cases, tablerow_tag.TablerowNode)

    def test_raw_tag(self):
        """Test that we can parse raw tags."""
        test_cases = [
            Case("raw literal", "{% raw %}foo{% endraw %}", "foo"),
            Case("raw object", "{% raw %}{{ foo }}{% endraw %}", r"{{ foo }}"),
            Case(
                "raw tag",
                "{% raw %}{% assign x = 1 %}{% endraw %}",
                "{% assign x = 1 %}",
            ),
        ]

        self._test(test_cases, literal.LiteralNode)

    def test_capture_tag(self):
        """Test that we can parse capture tags."""
        test_cases = [
            Case(
                "capture literal",
                "{% capture some %}foo{% endcapture %}",
                "var some = { foo }",
            ),
            Case(
                "capture literal and statement",
                "{% capture greeting %}Hello, {{ customer.first_name }}.{% endcapture %}",
                "var greeting = { Hello, `customer.first_name`. }",
            ),
        ]

        self._test(test_cases, capture_tag.CaptureNode)

    def test_cycle_tag(self):
        """Test that we can parse cycle tags."""
        test_cases = [
            Case(
                "cycle with no identifier",
                '{% cycle "some", "other" %}',
                "cycle ('some', 'other')",
            ),
            Case(
                "cycle with identifier",
                '{% cycle "foo": "some", "other" %}',
                "cycle ('foo': 'some', 'other')",
            ),
        ]

        self._test(test_cases, cycle_tag.CycleNode)

    def test_assign_tag(self):
        """Test that we can parse assign tags."""
        test_cases = [
            Case(
                "assign a string",
                '{% assign foo = "some" %}',
                "var (foo = 'some')",
            ),
        ]

        self._test(test_cases, assign_tag.AssignNode)

    def test_increment_tag(self):
        """Test that we can parse increment tags."""
        test_cases = [
            Case(
                "increment a named counter",
                "{% increment foo %}",
                "foo += 1",
            ),
        ]

        self._test(test_cases, increment_tag.IncrementNode)

    def test_decrement_tag(self):
        """Test that we can parse decrement tags."""
        test_cases = [
            Case(
                "decrement a named counter",
                "{% decrement foo %}",
                "foo -= 1",
            ),
        ]

        self._test(test_cases, decrement_tag.DecrementNode)

    def test_echo_tag(self):
        """Test that we can parse 'echo' tags."""
        test_cases = [
            Case("string literal", "{% echo 'hello' %}", "`'hello'`"),
            Case("identifier", "{% echo product %}", "`product`"),
            Case(
                "identifier with a filter",
                "{% echo product.title | upcase %}",
                "`product.title | upcase`",
            ),
        ]

        self._test(test_cases, echo_tag.EchoNode)

    def test_include_tag(self):
        """Test that we can parse 'include' tags."""
        test_cases = [
            Case(
                "simple string literal", "{% include 'product' %}", "include('product')"
            ),
            Case(
                "name from identifier", "{% include some.name %}", "include(some.name)"
            ),
            Case(
                "string literal with variable",
                "{% include 'product' with products[0] %}",
                "include('product' with products.0)",
            ),
            Case(
                "string literal with aliased variable",
                "{% include 'product' with products[0] as foo %}",
                "include('product' with products.0 as foo)",
            ),
            Case(
                "string literal with keyword argument",
                "{% include 'product', foo: 'bar' %}",
                "include('product', foo='bar')",
            ),
        ]

        self._test(test_cases, include_tag.IncludeNode)

    def test_render_tag(self):
        """Test that we can parse 'render' tags."""
        test_cases = [
            Case(
                "simple string literal", "{% render 'product' %}", "render('product')"
            ),
            Case(
                "string literal with variable",
                "{% render 'product' with products[0] %}",
                "render('product' with products.0)",
            ),
            Case(
                "string literal with aliased variable",
                "{% render 'product' with products[0] as foo %}",
                "render('product' with products.0 as foo)",
            ),
            Case(
                "string literal with keyword argument",
                "{% render 'product', foo: 'bar' %}",
                "render('product', foo='bar')",
            ),
        ]

        self._test(test_cases, render_tag.RenderNode)



from liquid.filter import Filter, no_args, number_required

# pylint: disable=arguments-differ, too-few-public-methods


class MoneyWithCurrency(Filter):

    __slots__ = ()

    name = "money_with_currency"

    @no_args
    @number_required
    def __call__(self, money):
        return f"$ {money / 100.0:.2f} USD"


class Money(Filter):

    __slots__ = ()

    name = "money"

    @no_args
    @number_required
    def __call__(self, money):
        return f"$ {money / 100.0:.2f}"



import unittest

from liquid.utils import LRUCache


class TemplateCacheTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cache = LRUCache(capacity=5)
        self.cache["foo"] = "bar"
        self.cache["some"] = "other"

    def test_copy(self):
        """Test that we can copy a cache."""
        copy = self.cache.copy()
        self.assertEqual(self.cache.items(), copy.items())

        copy["some"] = "different"
        self.assertNotEqual(self.cache.items(), copy.items())

    def test_set_default(self):
        """Test that we can set a default value for cache keys."""
        self.cache.setdefault("foo", "baz")
        self.assertEqual(self.cache["foo"], "bar")

        val = self.cache.setdefault("hello", "there")
        self.assertEqual(self.cache["hello"], val)

    def test_clear(self):
        """Test that we can clear all items from the cache."""
        self.assertEqual(len(self.cache), 2)
        self.cache.clear()
        self.assertEqual(len(self.cache), 0)

    def test_contains(self):
        """Test that we can check for membership."""
        self.assertEqual("foo" in self.cache, True)

    def test_delete(self):
        """Test that we can remove items from the cache."""
        self.assertEqual(len(self.cache), 2)
        del self.cache["foo"]
        self.assertEqual("foo" in self.cache, False)
        self.assertEqual(len(self.cache), 1)

    def test_values(self):
        """Test that we can get a list of cache values."""
        self.assertEqual(self.cache.values(), ["other", "bar"])

    def test_keys(self):
        """Test that we can get a list of cache keys."""
        self.assertEqual(self.cache.keys(), ["some", "foo"])

    def test_reversed(self):
        """Test that we can get a list of cache keys, oldest first."""
        self.assertEqual(list(reversed(self.cache)), ["foo", "some"])

    def test_capacity(self):
        """Test that the cache does not exceed capacity."""
        self.cache["a"] = 1
        self.cache["b"] = 2
        self.cache["c"] = 3
        self.assertEqual(len(self.cache), 5)

        self.cache["d"] = 4
        self.assertEqual(len(self.cache), 5)

    def test_priority(self):
        """Test that recently used items are not removed from the cache."""
        self.cache["a"] = 1
        self.cache["b"] = 2
        self.cache["c"] = 3
        self.assertEqual(len(self.cache), 5)

        self.assertEqual("foo" in self.cache, True)
        self.cache["d"] = 4
        self.assertEqual("foo" in self.cache, False)

        self.assertEqual("some" in self.cache, True)
        self.cache["some"] = "updated"
        self.cache["e"] = 5
        self.assertEqual("some" in self.cache, True)
        self.assertEqual("a" in self.cache, False)



"""Common parse tree nodes."""

from abc import ABC, abstractmethod
from io import StringIO

from typing import List
from typing import Optional
from typing import TextIO
from typing import Collection
from typing import Union
from typing import NoReturn

from liquid import __version__
from liquid import Context
from liquid import Expression

from liquid.token import Token, TOKEN_TAG

from liquid.exceptions import DisabledTagError
from liquid.exceptions import Error


class Node(ABC):
    """Base class for all nodes in a parse tree."""

    __slots__ = ()

    def token(self) -> Token:
        """The token that started this node."""
        return getattr(self, "tok", None)

    def raise_for_disabled(self, disabled_tags: Collection[str]):
        """Raise a DisabledTagError if this node's type is in the given list."""
        tok = self.token()
        if tok.type == TOKEN_TAG and tok.value in disabled_tags:
            raise DisabledTagError(
                f"{tok.value} usage is not allowed in this context",
                linenum=tok.linenum,
            )

    def render(self, context: Context, buffer: TextIO) -> Optional[bool]:
        """Check disabled tags before delegating to `render_to_output`."""
        if context.disabled_tags:
            self.raise_for_disabled(context.disabled_tags)
        return self.render_to_output(context, buffer)

    @abstractmethod
    def render_to_output(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Union[Optional[bool], NoReturn]:
        """Render this node to the output buffer."""


class ParseTree(Node):
    """The root node of all syntax trees."""

    __slots__ = ("statements", "version")

    def __init__(self):
        self.statements = []
        self.version = __version__

    def __str__(self):  # pragma: no cover
        return "".join(str(s) for s in self.statements)

    def __repr__(self):
        return f"ParseTree({self.statements})"

    def render_to_output(self, context: Context, buffer: TextIO):
        for stmt in self.statements:
            stmt.render(context, buffer)


class IllegalNode(Node):
    """Parse tree node representing an illegal or unregistered tag.

    Illegal nodes are necessary when running in "warn" or "lax" mode. In
    strict mode, an exception should be raised as soon as an illegal token
    is found.
    """

    __slots__ = ("tok",)

    def __init__(self, tok: Token):
        self.tok = tok

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return f"IllegalNode(tok={self.tok})"

    def render_to_output(self, context: Context, buffer: TextIO):
        pass


class BlockNode(Node):
    """A parse tree node representing a sequence of statements."""

    __slots__ = ("tok", "statements")

    def __init__(self, tok: Token, statements: Optional[List[Node]] = None):
        self.tok = tok
        self.statements = statements or []

    def __str__(self):
        return "".join(str(s) for s in self.statements)

    def render_to_output(self, context: Context, buffer: TextIO):
        # This intermediate buffer is used to suppress blocks that, when rendered,
        # contain only whitespace.
        buf = StringIO()

        for stmt in self.statements:
            try:
                stmt.render(context, buf)
            except Error as err:
                # Maybe resume rendering the block after an error.
                context.error(err)
            except Exception:
                # Write what we have so far and stop rendering the block.
                val = buf.getvalue()
                if not val.isspace():
                    buffer.write(val)
                raise

        # Don't write to the output buffer if the block contains only whitespace.
        val = buf.getvalue()
        if not val.isspace():
            buffer.write(val)


class ConditionalBlockNode(Node):
    """A parse tree node representing a sequence of statements and a conditional
    expression."""

    __slots__ = ("tok", "condition", "block")

    def __init__(self, tok: Token, condition: Expression, block: BlockNode = None):
        self.tok = tok
        self.condition = condition
        self.block = block

    def __str__(self):
        return f"{self.condition} {{ {self.block} }}"

    def render_to_output(self, context: Context, buffer: TextIO):
        if self.condition.evaluate(context):
            self.block.render(context, buffer)
            return True
        return False



import unittest
from inspect import isclass
from typing import NamedTuple, Any, List, Dict

from liquid.environment import Environment
from liquid.exceptions import FilterArgumentError

from liquid.builtin.filters.string import (
    Capitalize,
    Append,
    Downcase,
    Escape,
    EscapeOnce,
    LStrip,
    NewlineToBR,
    Prepend,
    Remove,
    RemoveFirst,
    Replace,
    ReplaceFirst,
    Slice,
    Split,
    Upcase,
    Strip,
    RStrip,
    StripHTML,
    StripNewlines,
    Truncate,
    TruncateWords,
    URLEncode,
    URLDecode,
)


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class StringFilterTestCase(unittest.TestCase):
    """Test string filter functions."""

    def _test(self, filter_cls, test_cases):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        func = filter_cls(env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, FilterArgumentError
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_capitalize(self):
        """Test capitalize filter function."""

        test_cases = [
            Case(
                description="lower case string",
                val="hello",
                args=[],
                kwargs={},
                expect="Hello",
            ),
            Case(
                description="already capitalized string",
                val="Hello",
                args=[],
                kwargs={},
                expect="Hello",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[2],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Capitalize, test_cases)

    def test_append(self):
        """Test append filter function."""

        test_cases = [
            Case(
                description="concat",
                val="hello",
                args=["there"],
                kwargs={},
                expect="hellothere",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="5there",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello5",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Append, test_cases)

    def test_downcase(self):
        """Test downcase filter function."""

        test_cases = [
            Case(
                description="make lower case",
                val="HELLO",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Downcase, test_cases)

    def test_escape(self):
        """Test escape filter function."""

        test_cases = [
            Case(
                description="make HTML-safe",
                val="<p>test</p>",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Escape, test_cases)

    def test_escape_once(self):
        """Test escape_once filter function."""

        test_cases = [
            Case(
                description="make HTML-safe",
                val="&lt;p&gt;test&lt;/p&gt;",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="make HTML-safe from mixed safe and markup.",
                val="&lt;p&gt;test&lt;/p&gt;<p>test</p>",
                args=[],
                kwargs={},
                expect="&lt;p&gt;test&lt;/p&gt;&lt;p&gt;test&lt;/p&gt;",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="HELLO",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(EscapeOnce, test_cases)

    def test_lstrip(self):
        """Test lstrip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello \t\r\n  ",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect="hello  \t\r\n ",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(LStrip, test_cases)

    def test_newline_to_br(self):
        """Test newline_to_br filter function."""

        test_cases = [
            Case(
                description="string with newlines",
                val="- apples\n- oranges\n",
                args=[],
                kwargs={},
                expect="- apples<br />- oranges<br />",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(NewlineToBR, test_cases)

    def test_prepend(self):
        """Test prepend filter function."""

        test_cases = [
            Case(
                description="concat",
                val="hello",
                args=["there"],
                kwargs={},
                expect="therehello",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="there5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="5hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Prepend, test_cases)

    def test_remove(self):
        """Test remove filter function."""

        test_cases = [
            Case(
                description="remove substrings",
                val="I strained to see the train through the rain",
                args=["rain"],
                kwargs={},
                expect="I sted to see the t through the ",
            ),
            Case(
                description="not a string",
                val=5,
                args=["there"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Remove, test_cases)

    def test_remove_first(self):
        """Test remove_first filter function."""

        test_cases = [
            Case(
                description="remove substrings",
                val="I strained to see the train through the rain",
                args=["rain"],
                kwargs={},
                expect="I sted to see the train through the rain",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(RemoveFirst, test_cases)

    def test_replace(self):
        """Test replace filter function."""

        test_cases = [
            Case(
                description="replace substrings",
                val="Take my protein pills and put my helmet on",
                args=["my", "your"],
                kwargs={},
                expect="Take your protein pills and put your helmet on",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain", "foo"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello",
                args=[5, "your"],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=["my"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Replace, test_cases)

    def test_replace_first(self):
        """Test replace_first filter function."""

        test_cases = [
            Case(
                description="replace substrings",
                val="Take my protein pills and put my helmet on",
                args=["my", "your"],
                kwargs={},
                expect="Take your protein pills and put my helmet on",
            ),
            Case(
                description="not a string",
                val=5,
                args=["rain", "foo"],
                kwargs={},
                expect="5",
            ),
            Case(
                description="argument not a string",
                val="hello5",
                args=[5, "your"],
                kwargs={},
                expect="helloyour",
            ),
            Case(
                description="missing argument",
                val="hello",
                args=["my"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=["how", "are", "you"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(ReplaceFirst, test_cases)

    def test_slice(self):
        """Test slice filter function."""

        test_cases = [
            Case(
                description="zero",
                val="hello",
                args=[0],
                kwargs={},
                expect="h",
            ),
            Case(
                description="one",
                val="hello",
                args=[1],
                kwargs={},
                expect="e",
            ),
            Case(
                description="one length three",
                val="hello",
                args=[1, 3],
                kwargs={},
                expect="ell",
            ),
            Case(
                description="out of range",
                val="hello",
                args=[99],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a string",
                val=5,
                args=[1],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="first argument not an integer",
                val="hello",
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="second argument not an integer",
                val="hello",
                args=[5, "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing arguments",
                val="hello",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[1, 2, 3],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="unexpected keyword arguments",
                val="hello",
                args=[1, 2],
                kwargs={"x": "y"},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Slice, test_cases)

    def test_split(self):
        """Test split filter function."""

        test_cases = [
            Case(
                description="split string",
                val="Hi, how are you today?",
                args=[
                    " ",
                ],
                kwargs={},
                expect=["Hi,", "how", "are", "you", "today?"],
            ),
            Case(
                description="not a string",
                val=5,
                args=[" "],
                kwargs={},
                expect=["5"],
            ),
            Case(
                description="argument not a string",
                val="hello th1ere",
                args=[1],
                kwargs={},
                expect=["hello th", "ere"],
            ),
            Case(
                description="missing argument",
                val="hello there",
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello there",
                args=[" ", ","],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Split, test_cases)

    def test_upcase(self):
        """Test upcase filter function."""

        test_cases = [
            Case(
                description="make lower case",
                val="hello",
                args=[],
                kwargs={},
                expect="HELLO",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Upcase, test_cases)

    def test_strip(self):
        """Test strip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Strip, test_cases)

    def test_rstrip(self):
        """Test rstrip filter function."""

        test_cases = [
            Case(
                description="left padded",
                val=" \t\r\n  hello",
                args=[],
                kwargs={},
                expect=" \t\r\n  hello",
            ),
            Case(
                description="right padded",
                val="hello \t\r\n  ",
                args=[],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="left and right padded",
                val=" \t\r\n  hello  \t\r\n ",
                args=[],
                kwargs={},
                expect=" \t\r\n  hello",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(RStrip, test_cases)

    def test_strip_html(self):
        """Test strip_html filter function."""

        test_cases = [
            Case(
                description="some HTML markup",
                val="Have <em>you</em> read <strong>Ulysses</strong> &amp; &#20;?",
                args=[],
                kwargs={},
                expect="Have you read Ulysses &amp; &#20;?",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(StripHTML, test_cases)

    def test_strip_newlines(self):
        """Test strip_newlines filter function."""

        test_cases = [
            Case(
                description="newline and other whitespace",
                val="hello there\nyou",
                args=[],
                kwargs={},
                expect="hello thereyou",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(StripNewlines, test_cases)

    def test_truncate_newlines(self):
        """Test truncate filter function."""

        test_cases = [
            Case(
                description="default end",
                val="Ground control to Major Tom.",
                args=[20],
                kwargs={},
                expect="Ground control to...",
            ),
            Case(
                description="custom end",
                val="Ground control to Major Tom.",
                args=[25, ", and so on"],
                kwargs={},
                expect="Ground control, and so on",
            ),
            Case(
                description="no end",
                val="Ground control to Major Tom.",
                args=[20, ""],
                kwargs={},
                expect="Ground control to Ma",
            ),
            Case(
                description="string is shorter than length",
                val="Ground control",
                args=[20],
                kwargs={},
                expect="Ground control",
            ),
            Case(
                description="not a string",
                val=5,
                args=[10],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[5, "foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Truncate, test_cases)

    def test_truncatewords_newlines(self):
        """Test truncatewords filter function."""

        test_cases = [
            Case(
                description="default end",
                val="Ground control to Major Tom.",
                args=[3],
                kwargs={},
                expect="Ground control to...",
            ),
            Case(
                description="custom end",
                val="Ground control to Major Tom.",
                args=[3, "--"],
                kwargs={},
                expect="Ground control to--",
            ),
            Case(
                description="no end",
                val="Ground control to Major Tom.",
                args=[3, ""],
                kwargs={},
                expect="Ground control to",
            ),
            Case(
                description="fewer words than word count",
                val="Ground control",
                args=[3],
                kwargs={},
                expect="Ground control",
            ),
            Case(
                description="not a string",
                val=5,
                args=[10],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="too many arguments",
                val="hello",
                args=[5, "foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(TruncateWords, test_cases)

    def test_url_encode_html(self):
        """Test url_encode filter function."""

        test_cases = [
            Case(
                description="some special URL characters",
                val="email address is bob@example.com!",
                args=[],
                kwargs={},
                expect="email+address+is+bob%40example.com%21",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(URLEncode, test_cases)

    def test_url_decode_html(self):
        """Test url_decode filter function."""

        test_cases = [
            Case(
                description="some special URL characters",
                val="email+address+is+bob%40example.com%21",
                args=[],
                kwargs={},
                expect="email address is bob@example.com!",
            ),
            Case(
                description="not a string",
                val=5,
                args=[],
                kwargs={},
                expect="5",
            ),
            Case(
                description="unexpected argument",
                val="hello",
                args=[5],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(URLDecode, test_cases)



from liquid.filter import (
    Filter,
    maybe_one_arg,
    no_args,
    one_maybe_two_args,
    string_required,
    two_string_args_required,
    maybe_one_arg_required,
    one_arg_required,
)

from liquid.exceptions import FilterArgumentError


TRANSLATION_TABLE = {"'": None, '"': None, "(": None, ")": None, "[": None, "]": None}


# pylint: disable=arguments-differ, too-few-public-methods


def to_handle(s):
    _s = s.lower().translate(TRANSLATION_TABLE)
    # _s = re.sub(r"\W+", "-", _s)
    _s = "-".join(_s.split())
    # if _s.endswith("-"):
    #     _s = _s[:-1]
    # if _s.startswith("-"):
    #     _s = s[1:]

    return _s


class AssetUrl(Filter):

    __slots__ = ()

    name = "asset_url"

    @no_args
    def __call__(self, val, *args, **kwargs):
        return f"/files/1/[shop_id]/[shop_id]/assets/{val}"


class GlobalAssetUrl(Filter):

    __slots__ = ()

    name = "global_asset_url"

    @no_args
    def __call__(self, val, *args, **kwargs):
        return f"/global/{val}"


class ShopifyAssetUrl(Filter):

    __slots__ = ()

    name = "shopify_asset_url"

    @no_args
    def __call__(self, val, *args, **kwargs):
        return f"/shopify/{val}"


class ScriptTag(Filter):

    __slots__ = ()

    name = "script_tag"

    @no_args
    @string_required
    def __call__(self, val, *args, **kwargs):
        return f'<script src="{val}" type="text/javascript"></script>'


class StylesheetTag(Filter):

    __slots__ = ()

    name = "stylesheet_tag"

    @string_required
    def __call__(self, val, *args, **kwargs):
        if maybe_one_arg(self.name, args, kwargs):
            media = args[0]
        else:
            media = "all"

        return (
            f'<link href="{val}" rel="stylesheet" type="text/css"  media="{media}"  />'
        )


def link_to(link, url, title=""):
    return f'<a href="{url}" title="{title}">{link}</a>'


class LinkTo(Filter):

    __slots__ = ()

    name = "link_to"

    def __call__(self, val, *args, **kwargs):
        url, title = one_maybe_two_args(self.name, args, kwargs)
        if not title:
            title = ""

        return link_to(val, url, title)


class Within(Filter):

    __slots__ = ()

    name = "within"

    @string_required
    @one_arg_required
    def __call__(self, url, collection):
        # XXX: Strip slash
        return f"/collections/{collection['handle']}/{url}"


class ImgTag(Filter):

    __slots__ = ()

    name = "img_tag"

    @string_required
    def __call__(self, val, *args, **kwargs):
        if maybe_one_arg(self.name, args, kwargs):
            alt = args[0]
        else:
            alt = ""

        return f'<img src="{val}" alt="{alt}" />'


class URLForVendor(Filter):

    __slots__ = ()

    name = "url_for_vendor"

    @string_required
    @no_args
    def __call__(self, val, *args, **kwargs):
        return f"/collections/{to_handle(val)}"


class URLForType(Filter):

    __slots__ = ()

    name = "url_for_type"

    @string_required
    @no_args
    def __call__(self, val, *args, **kwargs):
        return f"/collections/{to_handle(val)}"


class LinkToVendor(Filter):

    __slots__ = ()

    name = "link_to_vendor"

    @string_required
    @no_args
    def __call__(self, val, *args, **kwargs):
        vendor = val
        if not vendor:
            return "Unknown Vendor"

        url_for_vendor = f'"/collections/{to_handle(val)}"'
        return f'<a href="{url_for_vendor}" title="{vendor}">{vendor}</a>'


class LinkToType(Filter):

    __slots__ = ()

    name = "link_to_type"

    @string_required
    @no_args
    def __call__(self, val, *args, **kwargs):
        typ = val
        if not typ:
            return "Unknown Type"

        url_for_type = f"/collections/{to_handle(val)}"
        return f'<a href="{url_for_type}" title="{typ}">{typ}</a>'


class ProductImgURL(Filter):

    __slots__ = ()

    name = "product_img_url"

    @string_required
    @maybe_one_arg_required
    def __call__(self, url, style="small"):
        if not url.startswith("products/"):
            raise FilterArgumentError(
                f"{self.name}: can only be called on product images"
            )

        if style == "original":
            return f"/files/shops/random_number/{url}"

        if style in ("grande", "large", "medium", "compact", "small", "thumb", "icon"):
            # XXX: Does not attempt to mimic reference implementation URL.
            return f"/files/shops/random_number/products/{style}"

        raise FilterArgumentError(
            f"{self.name}: valid parameters for filter 'size' are: "
            "original, grande, large, medium, compact, small, thumb and icon"
        )


class DefaultPagination(Filter):

    __slots__ = ()

    name = "default_pagination"

    @no_args
    def __call__(self, paginate):
        html = []

        if paginate["previous"]:
            link = link_to(paginate["previous"]["title"], paginate["previous"]["url"])
            html.append(f'<span class="prev">{link}</span>')

        for part in paginate["parts"]:
            if part["is_link"]:
                link = link_to(part["title"], part["url"])
                html.append(f'<span class="page">{link}</span>')
            elif (
                part["title"] == paginate["current_page"]
            ):  # XXX: Leading integer comparison.
                title = part["title"]
                html.append(f'<span class="page current">{title}</span>')
            else:
                title = part["title"]
                html.append(f'<span class="deco">{title}</span>')

        if paginate["next"]:
            link = link_to(paginate["next"]["title"], paginate["next"]["url"])
            html.append(f'<span class="next">#{link}</span>')

        return " ".join(html)


class Pluralize(Filter):

    __slots__ = ()

    name = "pluralize"

    @two_string_args_required
    def __call__(self, val, *args, **kwargs):
        singular, plural = args[0], args[1]
        if val == 1:
            return singular
        return plural
